"""
Cannlytics Data Initialization | Cannlytics
Copyright (c) 2022 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 4/21/2022
Updated: 9/9/2022
"""
from .opendata.opendata import OpenData
from .data import (
    aggregate_datasets,
    create_hash,
    create_sample_id,
    find_first_value,
    parse_data_block,
    write_to_worksheet,
)

__all__ = [
    'OpenData',
    'aggregate_datasets',
    'create_hash',
    'create_sample_id',
    'find_first_value',
    'parse_data_block',
    'write_to_worksheet',
]



"""
Firebase Module | Cannlytics
Copyright (c) 2021-2023 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 2/7/2021
Updated: 1/10/2023
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: A wrapper of `firebase_admin` to make interacting with a Firestore
database and Firebase Storage buckets even easier. For more information, see
<https://firebase.google.com/docs/>.

Example:

```py
# Initialize Firebase with a .env file with `GOOGLE_APPLICATION_CREDENTIALS`
# pointing to the path of your service account.
database = initialize_firebase('./env')
```
"""
# Standard imports
from datetime import datetime, timedelta
from os import listdir
from os.path import isfile, join
from typing import Any, List, Optional, Tuple
from dotenv import dotenv_values

# External imports
import requests
import ulid
from firebase_admin import (
    auth,
    credentials,
    firestore,
    initialize_app,
    storage,
)
from google.cloud import secretmanager
from google.cloud.firestore import ArrayUnion, ArrayRemove, Increment
from google.cloud.firestore_v1.collection import CollectionReference
from google.cloud.firestore_v1.transforms import DELETE_FIELD
# FIXME: Raises the error below in Cloud Run.
# AttributeError: partially initialized module 'pandas' has no attribute 'core' (most likely due to a circular import)
try:
    from pandas import notnull, read_csv, read_excel, DataFrame, Series
except AttributeError:
    print('Pandas may not be installed.')

# Internal imports.
from cannlytics.utils import get_random_string, snake_case

# The maximum number of documents to include in batch updates.
# The official limit is 500, but pushing too close to the limit
# increases the likelihood of failure to commit.
MAX_BATCH_SIZE = 420


# === Core ===

def initialize_firebase(
        env_file: Optional[str] = None,
        key_path: Optional[str] = None,
        bucket_name: Optional[str] = None,
        project_id: Optional[str] = None,
    ):
    """Initialize Firebase, unless already initialized. Searches for environment
    credentials if `key_path` is not specified.
    Args:
        env_file (str): An .env file that contains `GOOGLE_APPLICATION_CREDENTIALS`.
        key_path (str): A path to your service account credentials (optional).
        project_id (str): A Firebase project ID (optional).
        bucket_name (str): A Cloud Storage bucket name (optional).
    Returns:
        (Firestore client): A Firestore database instance.
    """
    cred = None
    options = {}
    if env_file:
        config = dotenv_values(env_file)
        key_path = config['GOOGLE_APPLICATION_CREDENTIALS']
    elif key_path:
        cred = credentials.Certificate(key_path)
    if bucket_name:
        options['storageBucket'] = bucket_name.replace('gs://', '')
    if project_id:
        options['projectId'] = project_id
    try:
        initialize_app(cred, options)
    except ValueError:
        pass
    return firestore.client()


# === Firestore ===

def add_to_array(ref, field, value, database=None):
    """Add an element to a given field for a given reference.
    Args:
        ref (str): A document reference.
        field (str): A list field to create or update.
        value (dynamic): The value to be added to the list.
        database (Client): An optional existing Firestore database client.
    """
    if database is None:
        database = firestore.client()
    doc = create_reference(database, ref)
    doc.set({field: ArrayUnion([value])}, merge=True)


def create_document(ref, values, database=None):
    """Create a given document with given values, this leverages the
    same functionality as `update_document` thanks to `set` with `merge=True`.
    Args:
        ref (str): A document reference.
        values (str): A dictionary of values to update.
        database (Client): An optional existing Firestore database client.
    """
    update_document(ref, values, database)


def create_reference(database, path):
    """Create a database reference for a given path.
    Args:
        database (Firestore Client): The Firestore Client.
        path (str): The path to the document or collection.
    Returns:
        (ref): Either a document or collection reference.
    """
    ref = database
    parts = path.split('/')
    for index, part in enumerate(parts):
        if index % 2:
            ref = ref.document(part)
        else:
            ref = ref.collection(part)
    return ref


def delete_collection(ref, batch_size=420, database=None):
    """Delete a given collection, a batch at a time.
    Args:
        ref (str): A document reference.
        batch_size (int): The number of documents to delete at a time.
            The default is 420 and the maximum is 500.
        database (Client): An optional existing Firestore database client.
    """
    if database is None:
        database = firestore.client()
    col = create_reference(database, ref)
    docs = col.limit(batch_size).stream()
    deleted = 0
    for doc in docs:
        doc.reference.delete()
        deleted = deleted + 1
        if deleted >= batch_size:
            return delete_collection(col, batch_size)


def delete_document(ref: str, database=None):
    """Delete a given document.
    Args:
        ref (str): A document reference.
        database (Client): An optional existing Firestore database client.
    """
    if database is None:
        database = firestore.client()
    doc = create_reference(database, ref)
    doc.delete()


def delete_field(ref: str, field: str, database=None):
    """Delete a given field from a document.
    Args:
        ref (str): A document reference.
        field (str): The field to remove from the document.
        database (Client): An optional existing Firestore database client.
    """
    if database is None:
        database = firestore.client()
    doc = create_reference(database, ref)
    doc.set({field: DELETE_FIELD}, merge=True)


def remove_from_array(ref: str, field: str, value: Any, database=None):
    """Remove an element from a given field for a given reference.
    Args:
        ref (str): A document reference.
        field (str): A list field to update.
        value (dynamic): The value to be removed from the list.
        database (Client): An optional existing Firestore database client.
    """
    if database is None:
        database = firestore.client()
    doc = create_reference(database, ref)
    doc.set({field: ArrayRemove([value])}, merge=True)


def increment_value(ref: str, field: str, amount: int = 1, database=None):
    """Increment a given field for a given reference.
    Args:
        ref (str): A document reference.
        field (str): A numeric field to create or update.
        amount (int): The amount to increment, default 1.
        database (Client): An optional existing Firestore database client.
    """
    if database is None:
        database = firestore.client()
    doc = create_reference(database, ref)
    doc.set({field: Increment(amount)}, merge=True)


def update_document(ref: str, values: dict, database=None):
    """Update a given document with given values.
    Args:
        ref (str): A document reference.
        values (str): A dictionary of values to update.
        database (Client): An optional existing Firestore database client.
    """
    if database is None:
        database = firestore.client()
    doc = create_reference(database, ref)
    doc.set(values, merge=True)


def update_documents(refs: List[str], data: List[dict], database=None):
    """Batch update documents, up to the `MAX_BATCH_SIZE`, 420 by default.
    Args:
        refs (list): A list of document paths (str).
        data (list): A list of document data (dict).
        database (Client): An optional existing Firestore database client.
    """
    if database is None:
        database = firestore.client()
    ref_shards = [refs[i:i + MAX_BATCH_SIZE] for i in range(0, len(refs), MAX_BATCH_SIZE)]
    data_shards = [data[i:i + MAX_BATCH_SIZE] for i in range(0, len(data), MAX_BATCH_SIZE)]
    for shard_index, ref_shard in enumerate(ref_shards):
        data_shard = data_shards[shard_index]
        batch = database.batch()
        for index, ref in enumerate(ref_shard):
            values = data_shard[index]
            doc = create_reference(database, ref)
            batch.set(doc, values, merge=True)
        batch.commit()


def get_document(ref: str, database=None) -> dict:
    """Get a given document.
    Args:
        ref (str): A document reference.
        database (Client): A Firestore database client.
    Returns:
        (dict): Returns the document as a dictionary.
            Returns an empty dictionary if no data is found.
    """
    if database is None:
        database = firestore.client()
    doc = create_reference(database, ref)
    data = doc.get()
    try:
        values = data.to_dict()
        if values is None:
            return {}
        return {**{'id': data.id}, **values}
    except AttributeError:
        return {}


def get_collection( #pylint: disable=too-many-arguments
        ref: str,
        limit: int = None,
        order_by: str = None,
        desc: bool = False,
        # TODO: Allow user to pass a list of tuples.
        filters: List[dict] = None,
        database=None,
        start_at: dict = None,
) -> List[dict]:
    """Get documents from a collection.
    Args:
        ref (str): A document reference.
        limit (int): The maximum number of documents to return. The default is no limit.
        order_by (str): A field to order the documents by, with the default being none.
        desc (bool): The direction to order the documents by the order_by field.
        filters (list): Filters are dictionaries of the form
            `{'key': '', 'operation': '', 'value': ''}`.
            Filters apply [Firebase queries](https://firebase.google.com/docs/firestore/query-data/queries)
            to the given `key` for the given `value`.
            Operators include: `==`, `>=`, `<=`, `>`, `<`, `!=`,
            `in`, `not_in`, `array_contains`, `array_contains_any`.
        start_at (dict): Optional starting at value for pagination. Expect a dict
            of with `key` and `value` fields. For example: `{'key': 'number', 'value': 4 }`.
        database (Client): A Firestore database client.
    Returns:
        (list): A list of documents.
    """
    docs = []
    if database is None:
        database = firestore.client()
    collection = create_reference(database, ref)
    if filters is not None:
        for query_filter in filters:
            collection = collection.where(
                query_filter['key'], query_filter['operation'], query_filter['value']
            )
    if order_by and desc:
        collection = collection.order_by(order_by, direction='DESCENDING')
    elif order_by:
        collection = collection.order_by(order_by)
    if start_at:
        collection = collection.start_after({start_at['key']: start_at['value']})
    if limit:
        collection = collection.limit(limit)
    query = collection.stream()  # Only handles streams less than 2 mins.
    for doc in query:
        data = doc.to_dict()
        docs.append({**{'id': doc.id}, **data})
    return docs


def import_data(database, ref: str, data_file: str):
    """Import data into Firestore.
    Args:
        database (Firestore Client):
        ref (str): A collection or document reference.
        data_file (str): The path to the local data file to upload.
    !!! info "Wishlist"
        It would be desirable for the following functionality to be implemented:
        - Batch upload
        - Handle types <https://hackersandslackers.com/importing-excel-dates-times-into-pandas/>
    """
    try:
        data = read_csv(
            data_file,
            header=0,
            skip_blank_lines=True, 
            encoding='latin-1'
        )
    except:
        try:
            data = read_csv(data_file, sep=' ', header=None)
        except:
            try:
                data = read_csv(
                    data_file,
                    header=0,
                    skip_blank_lines=True, 
                    encoding='utf-16',
                    sep='\t',
                )
            except:
                data = read_excel(data_file, header=0)
    data.columns = map(snake_case, data.columns)
    data = data.where(notnull(data), None)
    data_ref = create_reference(database, ref)
    if isinstance(data_ref, CollectionReference):
        for index, values in data.iterrows():
            doc_id = str(index)
            doc_data = values.to_dict()
            data_ref.document(doc_id).set(doc_data, merge=True)
    else:
        doc_data = data.to_dict(orient='index')
        data_ref.set(doc_data, merge=True)


def export_data(database, ref: str, data_file: str):
    """Export data from Firestore.
    Args:
        database (Firestore Client):
        ref (str): A collection or document reference.
        data_file (str): The path to the local data file to upload.
    !!! info "Wishlist"
        Parse fields that are objects into fields, similar to below.
        ```py
        from pandas.io.json import json_normalize
        artist_and_track = json_normalize(
            data=tracks_response['tracks'],
            record_path='artists',
            meta=['id'],
            record_prefix='sp_artist_',
            meta_prefix='sp_track_',
            sep='_'
        )
        ```
    """
    data_ref = create_reference(database, ref)
    if isinstance(data_ref, CollectionReference):
        data = []
        docs = data_ref.stream()
        for doc in docs:
            doc_data = doc.to_dict()
            doc_data['id'] = doc.id
            data.append(doc_data)
        output = DataFrame(data)
    else:
        doc = data_ref.get()
        output = Series(doc.to_dict())
        output.name = doc.id
    if data_file.endswith('.csv'):
        output.to_csv(data_file)
    else:
        output.to_excel(data_file)


def create_id() -> str:
    """Generate a universal ID.
    Returns:
        (str): A unique lexicographic ID.
    """
    return ulid.new().str.lower()


def create_id_from_datetime(timestamp: datetime) -> str:
    """Create an ID from an existing datetime.
    Args:
        timestamp (datetime): The time to timestamp the ID.
    Returns:
        (str): A unique lexicographic ID.
    """
    return ulid.from_timestamp(timestamp).str.lower()


def get_id_timestamp(uid: str) -> datetime:
    """Get the datetime that an ID was created.
    Args:
        uid (str): A unique ID string.
    Returns:
        (datetime): The date when a unique lexicographic ID was generated.
    """
    return ulid.from_str(uid).timestamp().datetime


# === Authentication ===

def create_user(name: str, email: str) -> Tuple[str]:
    """
    Given user name and email, create an account.
    If the email is already being used, then nothing is returned.
    Args:
        name (str): A name for the user.
        email (str): The user's email.
    Returns:
        (tuple): Returns the User instance and a random password.
    """
    chars = 'abcdefghijklmnopqrstuvwxyz0123456789!@#$-_'
    password = get_random_string(42, chars)
    photo_url = f'https://robohash.org/{email}?set=set5'
    try:
        user = auth.create_user(
            uid=create_id(),
            email=email,
            email_verified=False,
            password=password,
            display_name=name,
            photo_url=photo_url,
            disabled=False,
        )
        return user, password
    except:
        return None, None


def create_custom_claims(
        uid: str ,
        email: str = None,
        claims: dict = None,
    ) -> None:
    """Create custom claims for a user to grant granular permission.
    The new custom claims will propagate to the user's ID token the
    next time a new one is issued.
    Args:
        uid (str): A user's ID.
        email (str): A user's email.
        claims (dict): A dictionary of the user's custom claims.
    """
    if email:
        user = auth.get_user_by_email(email)
        uid = user.uid
    auth.set_custom_user_claims(uid, claims)


def update_custom_claims(
        uid: str,
        email: str = None,
        claims: dict = None
    ) -> dict:
    """Update custom claims for a user.
    The new custom claims will propagate to the user's ID token the
    next time a new one is issued.
    Args:
        uid (str): A user's ID.
        email (str): A user's email.
        claims (dict): A dictionary of the user's custom claims.
    """
    if email:
        user = auth.get_user_by_email(email)
        uid = user.uid
    existing_claims = get_custom_claims(uid)
    if not existing_claims:
        existing_claims = {}
    # if existing_claims:
    #     existing_owner = existing_claims.get('owner', [])
    # else:
    #     existing_claims = {}
    #     existing_owner = []
    # current_owner = claims.get('owner', [])
    # FIXME: For now, only 1 organization per user.
    # claims['owner'] = list(set(existing_owner + current_owner))
    new_claims = {**existing_claims, **claims}
    auth.set_custom_user_claims(uid, new_claims)
    return new_claims


def get_custom_claims(name: str) -> dict:
    """Get custom claims for a user.
    Args:
        name (str): A user ID or user email.
    """
    user = get_user(name)
    return user.custom_claims


def create_custom_token(
        uid: Optional[str] = '',
        email: Optional[str] = None,
        claims: Optional[dict] = None
) -> bytes:
    """Create a custom token for a given user, expires after one hour.
    Args:
        uid (str): A user's ID.
        email (str): A user's email.
        claims (dict):  A dictionary of the user's claims.
    Returns:
        (bytes): A token minted from the input parameters.
    """
    if email:
        user = auth.get_user_by_email(email)
        uid = user.uid
    return auth.create_custom_token(uid, claims)


def create_session_cookie(
        id_token: str,
        expires_in: timedelta = None,
    ) -> bytes:
    """Create a session cookie.
    Args:
        id_token (str): A user ID token passed from the client.
        expires_in (timedelta): The time until the session will expire.
    Returns:
        (bytes): A session cookie in bytes.
    """
    if expires_in is None:
        expires_in = timedelta(days=7)
    return auth.create_session_cookie(id_token, expires_in=expires_in)


def revoke_refresh_tokens(token: str):
    """Revoke a user's refresh token.
    Args:
        token (str): The refresh token to authenticate a user.
    """
    auth.revoke_refresh_tokens(token)


def verify_token(token: str) -> dict:
    """Verify a user's custom token.
    Args:
        token (str): The custom token to authenticate a user.
    Returns:
        (dict): A dictionary of key-value pairs parsed from the decoded JWT.
    """
    return auth.verify_id_token(token)


def verify_session_cookie(
        session_cookie: str,
        check_revoked: bool = True,
        app=None,
    ) -> dict:
    """Verify a user's session cookie.
    Args:
        session_cookie (str): A session cookie to authenticate a user.
        check_revoked (bool): Checks if the cookie has been revoked or not (optional).
        app (App): A Firebase App instance.
    Returns:
        (dict): A dictionary of key-value pairs parsed from the decoded JWT.
    """
    return auth.verify_session_cookie(
        session_cookie,
        check_revoked=check_revoked,
        app=app,
    )


def get_user(name: str) -> Any:
    """Get a user by user ID or by email.
    Args:
        name (str): A user ID, email, or phone number.
    Returns:
        (UserRecord): A Firebase user object.
    """
    user = None
    try:
        user = auth.get_user(name)
    except:
        pass
    if user is None:
        try:
            user = auth.get_user_by_email(name)
        except:
            pass
    if user is None:
        try:
            user = auth.get_user_by_phone_number(name)
        except:
            pass
    return user


def get_users() -> list:
    """Get all Firebase users.
    Returns:
        (list): A list of Firebase users.
    """
    users = []
    for user in auth.list_users().iterate_all():
        users.append(user)
    return users


def update_user(existing_user: Any, data: dict):
    """Update a user.
    Args:
        existing_user (Firebase user): A Firebase user object.
        data (dict): The values of the user to update, which can include
            email, phone_number, email_verified, display_name, photo_url,
            and disabled.
    """
    values = {}
    fields = ['email', 'phone_number', 'email_verified', 'display_name',
              'photo_url', 'disabled']
    for field in fields:
        new_value = data.get(field)
        if new_value:
            values[field] = new_value
        else:
            values[field] = getattr(existing_user, field)
    return auth.update_user(
        existing_user.uid,
        email=values['email'],
        phone_number=values['phone_number'],
        email_verified=values['email_verified'],
        display_name=values['display_name'],
        photo_url=values['photo_url'],
        disabled=values['disabled'],
    )


def delete_user(uid: str):
    """Delete a user from Firebase.
    Args:
        uid (str): A user's ID.
    """
    auth.delete_user(uid)


def generate_password_reset_link(email: str) -> str:
    """Get a password reset link for a user given their email.
    Args:
        email (str): A user's email.
    Returns:
        (str): A password reset link for the user.
    """
    return auth.generate_password_reset_link(email)


# === Secret Management ===

def create_secret(project_id: str, secret_id: str, secret: Any) -> str:
    """Create a new secret with the given name. A secret is a logical wrapper
    around a collection of secret versions. Secret versions hold the actual
    secret material. 'Secret Manager Admin' permissions needed for service account.
    See <https://cloud.google.com/secret-manager/docs/creating-and-accessing-secrets>.
    Args:
        project_id (str): The project associated with the secret.
        secret_id (str): An ID for the secret.
        secret (dynamic): The secret data, a string or a dict can both be used.
    Returns:
        (str): The name of the created secret.
    """
    client = secretmanager.SecretManagerServiceClient()
    response = client.create_secret(
        parent=f'projects/{project_id}',
        secret_id=secret_id,
        secret=secret,
    )
    return response.name


def add_secret_version(
        project_id: str,
        secret_id: str,
        payload: Any,
    ) -> str:
    """
    Add a new secret version to the given secret with the provided payload.
    A secret version contains the actual contents of a secret.
    A secret version can be enabled, disabled, or destroyed.
    To change the contents of a secret, you create a new version.
    Adding a secret version requires the Secret Manager Admin role
    (roles/secretmanager.admin) on the secret, project, folder, or organization.
    Roles can't be granted on a secret version.
    'Secret Manager Admin' permissions needed for service account.
    See <https://cloud.google.com/secret-manager/docs/creating-and-accessing-secrets>.
    Args:
        project_id (str): A Firestore project ID.
        secret_id (str): An ID for the secret.
        payload (dynamic): The secret.
    Returns:
        (str): The secret version's name.
    """
    client = secretmanager.SecretManagerServiceClient()
    parent = f'projects/{project_id}/secrets/{secret_id}'
    payload = payload.encode('UTF-8')
    response = client.add_secret_version(
        parent=parent,
        payload={'data': payload},
    )
    return response.name


def access_secret_version(
        project_id: str,
        secret_id: str,
        version_id: Optional[str] = 'latest',
    ) -> str:
    """
    Access the payload for a given secret version if one exists. The version
    can be a version number as a string (e.g. "5") or an alias (e.g. "latest").
    'Secret Manager Admin' permissions needed for service account.
    See <https://cloud.google.com/secret-manager/docs/creating-and-accessing-secrets>.
    !!! Warning
        Do not print the secret in a production environment.
    Args:
        project_id (str): A Firestore project ID.
        secret_id (str): An ID for the secret.
        version_id (str): A version for the secret.
    Returns:
        (str): The secret value.
    """
    client = secretmanager.SecretManagerServiceClient()
    name = f'projects/{project_id}/secrets/{secret_id}/versions/{version_id}'
    response = client.access_secret_version({'name': name})
    return response.payload.data.decode('UTF-8')


# === Storage ===

def create_short_url(
        api_key: str,
        long_url: str,
        project_name: str,
    ) -> str:
    """Create a short URL to a specified file.
    Args:
        api_key (str): An API key for Firebase dynamic links.
        long_url (str): A URL to create a short, dynamic link.
        project_name (str): The name of the Firebase project.
    Returns:
        (str): A short link to the given URL.
    """
    try:
        url = f'https://firebasedynamiclinks.googleapis.com/v1/shortLinks?key={api_key}'
        data= {
            'dynamicLinkInfo': {
                'domainUriPrefix': f'https://{project_name}.page.link',
                'link': long_url,
            },
            # TODO: Make suffix a argument.
            'suffix': {'option': 'UNGUESSABLE'}
        }
        response = requests.post(url, json=data)
        return response.json()['shortLink']
    except ConnectionError:
        raise ConnectionError # Optional: Handle connection errors more elegantly?


def download_file(
        source_blob_name: str,
        destination_file_name: str,
        bucket_name: Optional[str] = None,
    ):
    """Downloads a file from Firebase Storage.
    Args:
        source_blob_name (str): The file name to upload.
        destination_file_name (str): The destination file name.
        bucket_name (str): The name of the storage bucket (optional).
    """
    bucket = storage.bucket(name=bucket_name)
    blob = bucket.blob(source_blob_name)
    blob.download_to_filename(destination_file_name)


def download_files(
        bucket_folder: str,
        local_folder: str,
        bucket_name: Optional[str] = None,
    ):
    """Download all files in a given Firebase Storage folder.
    Args:
        bucket_folder (str): A folder in the storage bucket.
        local_folder (str): The local folder to download files.
        bucket_name (str): The name of the storage bucket (optional).
    """
    bucket = storage.bucket(name=bucket_name)
    file_list = list_files(bucket_name, bucket_folder)
    for file in file_list:
        blob = bucket.blob(file)
        file_name = blob.name.split('/')[-1]
        blob.download_to_filename(local_folder + '/' + file_name)


def get_file_url(
        ref: str,
        bucket_name: Optional[str] = None,
        expiration=None,
    ) -> str:
    """Return the storage URL of a given file reference.
    Args:
        ref (str): The storage location of the file.
        bucket_name (str): The name of the storage bucket.
        expiration (datetime): The date for when the file link should expire.
    Returns:
        (str): The storage URL of the file.
    """
    if expiration is None:
        expiration = datetime.now() + timedelta(days=100 * 365)
    bucket = storage.bucket(name=bucket_name)
    blob = bucket.blob(ref)
    return blob.generate_signed_url(expiration)


def upload_file(
        destination_blob_name: str,
        source_file_name: Optional[str] = None,
        data_url: Optional[str] = None,
        content_type: Optional[str] = 'image/jpg',
        bucket_name: Optional[str] = None,
):
    """Upload file to Firebase Storage.
    Args:
        destination_blob_name (str): The name to save the file as.
        source_file_name (str): The local file name.
        data_url (str): The data URL to upload from a string.
        content_type (str): The content type of the file, when uploading from a string.
        bucket_name (str): The name of the storage bucket (optional).
    """
    bucket = storage.bucket(name=bucket_name)
    blob = bucket.blob(destination_blob_name)
    if source_file_name:
        blob.upload_from_filename(source_file_name)
    else:
        blob.upload_from_string(data_url, content_type=content_type)


def upload_files(
        bucket_folder: str,
        local_folder: str,
        bucket_name: Optional[str] = None,
    ):
    """Upload multiple files to Firebase Storage.
    Args:
        bucket_folder (str): A folder in the storage bucket to upload files.
        local_folder (str): The local folder of files to upload.
        bucket_name (str): The name of the storage bucket (optional).
    """
    bucket = storage.bucket(name=bucket_name)
    files = [f for f in listdir(local_folder) if isfile(join(local_folder, f))]
    for file in files:
        local_file = join(local_folder, file)
        blob = bucket.blob(bucket_folder + '/' + file)
        blob.upload_from_filename(local_file)


def list_files(
        bucket_folder: str,
        bucket_name: Optional[str] = None,
    ) -> List[str]:
    """List all files in GCP bucket folder.
    Args:
        bucket_folder (str): A folder in the storage bucket to list files.
        bucket_name (str): The name of the storage bucket (optional).
    Returns:
        (list): A list of file names in the given bucket.
    """
    bucket = storage.bucket(name=bucket_name)
    files = bucket.list_blobs(prefix=bucket_folder)
    return [file.name for file in files if '.' in file.name]


def delete_file(blob_name: str, bucket_name: Optional[str] = None):
    """Delete file from GCP bucket.
    Args:
        blob_name (str): The name of the file to delete.
        bucket_name (str): The name of the storage bucket (optional).
    """
    bucket = storage.bucket(name=bucket_name)
    bucket.delete_blob(blob_name)


def rename_file(
        bucket_folder: str,
        file_name: str,
        newfile_name: str,
        bucket_name: Optional[str] = None,
):
    """Rename file in GCP bucket.
    Args:
        bucket_folder (str): A folder in the storage bucket.
        file_name (str): The name of the file to rename.
        newfile_name (str): The new name for the file.
        bucket_name (str): The name of the storage bucket (optional).
    """
    bucket = storage.bucket(name=bucket_name)
    blob = bucket.blob(bucket_folder + '/' + file_name)
    bucket.rename_blob(blob, new_name=newfile_name)


# === Misc===

def create_log( #pylint: disable=too-many-arguments
        ref: str,
        claims: dict,
        action: str,
        log_type: str,
        key: str,
        changes: Any = None,
):
    """Create an activity log.
    Args:
        ref (str): Path to a collection of logs.
        claims (dict): A dict with user fields or a Firestore user object.
        action (str): The activity that took place.
        log_type (str): The log type.
        key (str): A key to recognize the action.
        changes (list): An optional list of changes that took place.
    """
    now = datetime.now()
    timestamp = datetime.now().isoformat()
    log_id = now.strftime('%Y-%m-%d_%H-%M-%S')
    log_entry = {
        'action': action,
        'type': log_type,
        'key': key,
        'created_at': timestamp,
        'log_id': log_id,
        'user': claims.get('uid'),
        'user_name': claims.get('display_name'),
        'user_email': claims.get('email'),
        'user_photo_url': claims.get('photo_url'),
        'changes': changes,
    }
    update_document(f'{ref}/{log_id}', log_entry)



"""
Transfer Model | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Transfer data model.
"""
# Standard imports.
from dataclasses import dataclass
from datetime import datetime

# Internal imports.
from .base import Model


@dataclass
class Transfer(Model):
    """A group of samples or inventory being transferred between
    two organizations."""
    _collection = 'organizations/%s/transfers'
    status: str = ''
    departed_at: datetime = datetime.now()
    arrived_at: datetime = datetime.now()
    transfer_type: str = ''
    sample_count: int = 0
    sender: str = ''
    sender_org_id: str = ''
    receiver: str = ''
    receiver_org_id: str = ''
    transporter: str = ''



"""
Cannlytics GIS Data Initialization | Cannlytics
Copyright (c) 2023 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 7/2/2023
Updated: 7/2/2023
"""
try:
    from .gis import (
        get_google_maps_api_key,
        get_state_data,
        get_state_population,
        geocode_addresses,
        search_for_address,
        get_transfer_distance,
        get_transfer_route,
        initialize_googlemaps,
    )

    __all__ = [
        get_google_maps_api_key,
        get_state_data,
        get_state_population,
        geocode_addresses,
        search_for_address,
        get_transfer_distance,
        get_transfer_route,
        initialize_googlemaps,
    ]
except ImportError:
    pass



"""
Cannlytics Metrc Client Initialization | Cannlytics
Copyright (c) 2021-2022 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/6/2021
Updated: 1/14/2023
"""
from typing import Any, Optional
from .client import Metrc
from .exceptions import MetrcAPIError
from .models import (
    Delivery,
    Category,
    Employee,
    Facility,
    Item,
    Location,
    Harvest,
    Package,
    Patient,
    Plant,
    PlantBatch,
    LabResult,
    Receipt,
    Strain,
    Transfer,
    TransferTemplate,
    Transaction,
    Waste,
)


def initialize_metrc(
        vendor_api_key: str,
        user_api_key: str,
        logs: Optional[bool] = False,
        test: Optional[bool] = False,
        primary_license: Optional[str] = '',
        state: Optional[str] = 'ca',
        client_class: Any = Metrc,
    ) -> Metrc:
    """This is a shortcut function which instantiates a Metrc
    client using a user API key and the vendor API key.
    Args:
        vendor_api_key (str): The vendor's API key.
        user_api_key (str): The user's API key.
        primary_license (str): An optional primary license to use if no license is specified.
        state (str): The state of the traceability system, `ca` by default.
        client_class: By default :class:`cannlytics.metrc.client.Client` is used.
    Returns:
        (Metrc): Returns an instance of the Metrc client.
    """
    return client_class(
        vendor_api_key,
        user_api_key,
        logs=logs,
        test=test,
        primary_license=primary_license,
        state=state,
    )

__all__ = [
    initialize_metrc,
    Metrc,
    MetrcAPIError,
    Delivery,
    Category,
    Employee,
    Facility,
    Item,
    Location,
    Harvest,
    Package,
    Patient,
    Plant,
    PlantBatch,
    LabResult,
    Receipt,
    Strain,
    Transfer,
    TransferTemplate,
    Transaction,
    Waste,
]



"""
Data Models | Cannlytics Engine
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 5/8/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Base data model schema for Cannlytics data models.
"""

# Standard imports.
from dataclasses import dataclass
from datetime import datetime

# External imports.
# pylint: disable=no-member
import ulid


@dataclass
class Model:
    """Base data model schema for Cannlytics data models."""
    created_at: datetime = datetime.now()
    updated_at: datetime = datetime.now()
    uid: str = ulid.new().str.lower()
    ref: str = ''

    @classmethod
    def from_dict(cls, data):
        """Initiate a document from a dictionary."""

        # Split the kwargs into native ones and new ones.
        native_args, new_args = {}, {}
        for key, value in data.items():
            if key in cls.__annotations__:
                native_args[key] = value
            else:
                new_args[key] = value

        # Use the native arguments to create the class.
        ret = cls(**native_args)

        # Add the arguments to the model.
        for new_name, new_val in new_args.items():
            setattr(ret, new_name, new_val)

        return ret

    def to_dict(self) -> dict:
        """Returns the model's properties as a dictionary."""
        return vars(self).copy()



"""
Certificate Model | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Certificate data model.
"""
# Standard imports.
from dataclasses import dataclass

# Internal imports.
from .base import Model


@dataclass
class Certificate(Model):
    """A certificate displaying the final results for analyses of a
    sample, a CoA. The CoA consists of a template and a reference to the
    generated PDF."""
    _collection = 'organizations/%s/certificates'
    template_storage_ref: str = ''
    template_version: str = ''

    def approve(self):
        """Approve a CoA after it has been reviewed, signing the
        approval line on the CoA. Approval requires `qa` claims.
        The CoA reference is updated with the signed CoA version."""
        return NotImplementedError

    def create_pdf(self):
        """Create a PDF representation of the CoA, with blank review
        and approve lines. """
        return NotImplementedError

    def review(self):
        """Review a certificate after it has been created. The CoA
        reference is updated with the signed CoA version."""
        return NotImplementedError



"""
Massachusetts Cannabis Control Commission (CCC) Open Data API Wrapper
Copyright (c) 2022 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 4/5/2022
Updated: 7/31/2022
License: MIT License <https://opensource.org/licenses/MIT>

Data sources:

    - Massachusetts Cannabis Control Commission Data Catalog
    https://masscannabiscontrol.com/open-data/data-catalog/

TODO:

    - [ ] Create a data guide.
    - [ ] Work on the dynamics of the column types.

FIXME: SQL queries do not appear to work.

"""
# Standard imports.
import json
import os
from typing import Any, Optional

# External imports.
import pandas as pd
from requests import Session

# Internal imports.
from cannlytics.utils.constants import DEFAULT_HEADERS


OPENDATA_ENDPOINTS = {
    'agent-gender-stats': 'hhjg-atjk',
    'agent-ethnicity-stats': 'pt2c-wb44',
    'licensees': 'albs-all',
    'licensees-approved': 'hmwt-yiqy',
    'licensees-pending': 'piib-tj3f',
    'licensees-demographics': '5dkg-e39p',
    'licensees-under-review-stats': 'pebi-jpc4',
    'licensees-application-stats': 'n6qz-us6r',
    'retail-sales-stats': '87rp-xn9v',
    'retail-sales-weekly': 'dt9b-i6ds',
    'retail-price-per-ounce': 'rqtv-uenj',
    'medical-stats': 'g5mj-5pg3',
    'plants': 'meau-plav',
    'sales': 'fren-z7jq',
}
OPENDATA_BOOLEAN_COLUMNS = [
    'activity_date',
    'app_create_date',
    'facilityisexpired',
    'facilityismedical',
    'facilityisretail',
    'facilityiscultivation',
    'facilityisdispensaryorstore',
    'facilityisinfusedmanufacturer',
    # is_abutters_notified
    'not_a_dbe',
    'priority',
]
OPENDATA_DATETIME_COLUMNS = [
    'activitysummarydate',
    'sale_period',
    'saledate',
]
OPENDATA_NUMERIC_COLUMNS = [
    'abutters_count',
    'application_fee',
    'average_spent',
    'countbasedtotal',
    'dollarcountbasedtotal',
    'dollarweightbasedtotal',
    'gross_sales',
    'harvestactivecount',
    'latitude',
    'lic_fee_amount',
    'longitude',
    'percent_total',
    'plantdestroyedcount',
    'plantfloweringcount',
    'plantharvestedcount',
    'plantvegetativecount',
    'price_per_ounce',
    'quantity',
    'square_footage_establishment',
    'strainactivecount',
    'total',
    'total_dollars',
    'totalprice',
    'total_units',
    'units',
    'weightbasedtotal',
]
OPENDATA_CODINGS = {
    '= 1 oz': 'price_per_ounce',
}


class APIError(Exception):
    """A primary error raised by the Open Data API."""

    def __init__(self, response):
        message = self.get_response_messages(response)
        super().__init__(message)
        self.response = response

    def get_response_messages(self, response):
        """Extract error messages from a Open Data API response.
        Args:
            response (Response): A request response from the Open Data API.
        Returns:
            (str): Returns any error messages.
        """
        try:
            return response.json()
        except:
            return response.text


class OpenData(object):
    """An instance of this class communicates with the
    Cannabis Control Commission of the Commonwealth of Massachusetts'
    Open Data catalog."""

    def __init__(self, headers=None) -> None:
        """Initialize an Open Data API client.
        Args:
            headers (dict): Headers for HTTP requests (optional).
        """
        self.base = 'https://masscannabiscontrol.com/resource/'
        self.endpoints = OPENDATA_ENDPOINTS
        if headers is None:
            self.headers = DEFAULT_HEADERS
        else:
            self.headers = headers
        self.session = Session()

    def get(self, endpoint:str, params: Optional[dict] = None) -> Any:
        """Make a request to the API.
        Args:
            endpoint (str): The API endpoint.
            params (dict): A dictionary of parameters.
        Returns:
            (DataFrame): The HTTP data formatted as a DataFrame.
        """
        url = os.path.join(self.base, f'{endpoint}.json')
        try:
            response = self.session.get(url, headers=self.headers, params=params)
        except ConnectionError:
            self.session = Session()
            response = self.session.get(url, headers=self.headers, params=params)
        if response.status_code != 200:
            raise APIError(response)
        try:
            body = response.json()
        except:
            body = json.loads(response.text.replace('\r', '').replace('\ufeff', ''))
        data = pd.DataFrame(body)
        data.columns = map(str.lower, data.columns)
        data.columns = [x.replace('$', 'dollars').replace('%', 'percent') for x in data.columns]
        data.rename(columns=OPENDATA_CODINGS, inplace=True)
        for key in list(set(OPENDATA_BOOLEAN_COLUMNS).intersection(data.columns)):
            data[key] = data[key].astype(bool, errors='ignore')
        for key in list(set(OPENDATA_DATETIME_COLUMNS).intersection(data.columns)):
            data[key] = pd.to_datetime(data[key], errors='ignore')
        for key in list(set(OPENDATA_NUMERIC_COLUMNS).intersection(data.columns)):
            data[key] = data[key].astype(float, errors='ignore')
        return data

    def get_agents(self, dataset: Optional[str] = 'gender-stats') -> Any:
        """Get agent statistics.
        Args:
            dataset (str): An optional dataset filter:
                * `gender-stats` (default)
                * `ethnicity-stats`
        Returns:
            (DataFrame): The agent data formatted as a DataFrame.
        """
        key = 'agent'
        if dataset:
            key += '-' + dataset
        endpoint = self.endpoints[key]
        return self.get(endpoint)

    def get_licensees(
            self,
            dataset: Optional[str] = '',
            limit: Optional[int] = 10_000,
            order_by: Optional[str] = 'app_create_date',
            ascending: Optional[bool] = False,
        ) -> Any:
        """Get Massachusetts licensee data and statistics.
        Args:
            dataset (str): An optional dataset filter:
                * `approved`
                * `pending`
                * `demographics`
                * `under-review-stats`
                * `application-stats`
            limit (int): A limit to the number of returned observations.
            order_by (str): The field to order the results, `app_create_date` by default.
            ascending (bool): If ordering results, ascending or descending. Descending by default.
        Returns:
            (DataFrame): The licensees data formatted as a DataFrame.
        """
        key = 'licensees'
        if dataset:
            key += '-' + dataset
        endpoint = self.endpoints[key]
        params = {
            '$limit': limit,
            '$order': order_by,
        }
        if not ascending:
            params['$order'] += ' DESC'
        return self.get(endpoint, params=params)

    def get_retail(
            self,
            dataset: Optional[str] = 'sales-stats',
            limit: Optional[int] = 10_000,
            order_by: Optional[str] = 'date',
            ascending: Optional[bool] = False,
        ) -> Any:
        """Get Massachusetts retail data and statistics.
        Args:
            dataset (str): An optional dataset filter:
                * `sales-stats` (default)
                * `sales-weekly`
                * `price-per-ounce`
            limit (int): A limit to the number of returned observations.
            order_by (str): The field to order the results, `app_create_date` by default.
            ascending (bool): If ordering results, ascending or descending. Descending by default.
        Returns:
            (DataFrame): The retail data formatted as a DataFrame.
        """
        key = 'retail'
        key += '-' + dataset
        endpoint = self.endpoints[key]
        params = {
            '$limit': limit,
            '$order': order_by,
        }
        if not ascending:
            params['$order'] += ' DESC'
        return self.get(endpoint, params=params)

    def get_medical(self, dataset: Optional[str] = 'stats') -> Any:
        """Get Massachusetts medical stats.
        Args:
            dataset (str): The specific medical dataset to retrieve.
        Returns:
            (DataFrame): The medical data formatted as a DataFrame.
        """
        key = 'medical'
        key += '-' + dataset
        endpoint = self.endpoints[key]
        return self.get(endpoint)

    def get_plants(
            self,
            limit: Optional[int] = 10_000,
            order_by: Optional[str] = 'activitysummarydate',
            ascending: Optional[bool] = False,
        ) -> Any:
        """Get Massachusetts cultivation data and statistics.
        Args:
            limit (int): A limit to the number of returned observations.
            order_by (str): The field to order the results, `app_create_date` by default.
            ascending (bool): If ordering results, ascending or descending. Descending by default.
        Returns:
            (DataFrame): The plant data formatted as a DataFrame.
        """
        key = 'plants'
        endpoint = self.endpoints[key]
        params = {
            '$limit': limit,
            '$order': order_by,
        }
        if not ascending:
            params['$order'] += ' DESC'
        return self.get(endpoint, params=params)

    def get_sales(
            self,
            limit: Optional[int] = 10_000,
            order_by: Optional[str] = 'activitysummarydate',
            ascending: Optional[bool] = False,
        ) -> Any:
        """Get Massachusetts sales data.
        Args:
            limit (int): A limit to the number of returned observations.
            order_by (str): The field to order the results, `app_create_date` by default.
            ascending (bool): If ordering results, ascending or descending. Descending by default.
        Returns:
            (DataFrame): The sales data formatted as a DataFrame.
        """
        key = 'sales'
        endpoint = self.endpoints[key]
        params = {
            '$limit': limit,
            '$order': order_by,
        }
        if not ascending:
            params['$order'] += ' DESC'
        return self.get(endpoint, params=params)



"""
Web Data Tools | Cannlytics
Copyright (c) 2021-2022 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 1/10/2021
Updated: 10/8/2022
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Resources:
    https://stackoverflow.com/questions/54416896/how-to-scrape-email-and-phone-numbers-from-a-list-of-websites
    https://hackersandslackers.com/scraping-urls-with-beautifulsoup/

TODO:
    Improve with requests-html - https://github.com/psf/requests-html
    - Get #about
    - Get absolute URLs
    - Search for text (prices/analyses)
        r.html.search('Python is a {} language')[0]
"""
import re
from typing import Any, Optional, Tuple
import requests
from bs4 import BeautifulSoup


# === General Web Scraping Tools ===

def format_params(parameters, **kwargs):
    """Format given keyword arguments HTTP request parameters.
    Returns:
        (dict): Returns the parameters as a dictionary.
    """
    params = {}
    for param in kwargs:
        if kwargs[param]:
            key = parameters[param]
            params[key] = kwargs[param]
    return params


def get_page_metadata(url: str) -> Tuple:
    """Get the metadata of a web page.
    Args:
        url (str): The URL to scrape.
    Returns:
        (HTTPResponse): The HTTP response.
        (str): The HTML text.
        (dict): A dictionary of metadata, including: `description`, `image_url`,
            `favicon`, and `brand_color`.
    """
    headers = {
        'Access-Control-Allow-Origin': '*',
        'Access-Control-Allow-Methods': 'GET',
        'Access-Control-Allow-Headers': 'Content-Type',
        'Access-Control-Max-Age': '3600',
        'User-Agent': 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:52.0) Gecko/20100101 Firefox/52.0',
    }
    # Handle URLs without http beginning
    if not url.startswith('http'):
        url = 'http://' + url
    response = requests.get(url, headers=headers)
    html = BeautifulSoup(response.content, 'html.parser')
    metadata = {
        'description': get_page_description(html),
        'image_url': get_page_image(html),  # FIXME: Append URL if relative path.
        'favicon': get_page_favicon(html, url),
        'brand_color': get_page_theme_color(html),
    }
    return response, html, metadata


def get_page_description(html: str) -> str:
    """Get the description of a web page.
    Args:
        html (str): A body of HTML text.
    Returns:
        (str): A description excerpt from the page.
    """
    description = None
    if html.find('meta', property='description'):
        description = html.find('meta', property='description').get('content')
    elif html.find('meta', property='og:description'):
        description = html.find('meta', property='og:description').get('content')
    elif html.find('meta', property='twitter:description'):
        description = html.find('meta', property='twitter:description').get('content')
    elif html.find('p'):
        description = html.find('p').contents
    if isinstance(description, list):
        try:
            description = description[0]
        except IndexError:
            pass
    return description


def get_page_image(html: str, index: Optional[int] = 0) -> str:
    """Get an image on a web page, the first image by default.
    Args:
        html (str): A body of HTML text.
    Returns:
        (str): Returns the first image URL if found.
    """
    image = None
    if html.find('meta', property='image'):
        image = html.find('meta', property='image').get('content')
    elif html.find('meta', property='og:image'):
        image = html.find('meta', property='og:image').get('content')
    elif html.find('meta', property='twitter:image'):
        image = html.find('meta', property='twitter:image').get('content')
    elif html.find('img', src=True):
        image = html.find_all('img')[index].get('src')
    return image


def get_page_favicon(html: str, url: Optional[str] = '') -> str:
    """Get the favicon from a web page.
    Args:
        html (str): A body of HTML text.
        url (str): The URL of the page.
    Returns:
        (str): The URL of any potential favicon.
    """
    if html.find('link', attrs={'rel': 'icon'}):
        favicon = html.find('link', attrs={'rel': 'icon'}).get('href')
    elif html.find('link', attrs={'rel': 'shortcut icon'}):
        favicon = html.find('link', attrs={'rel': 'shortcut icon'}).get('href')
    else:
        favicon = f"{url.rstrip('/')}/favicon.ico"
    return favicon


def get_page_theme_color(html: str) -> str:
    """Get the theme color of a web page.
    Args:
        html (str): A body of HTML text.
    Returns:
        (str): An hex color code if found.
    """
    if html.find('meta', property='theme-color'):
        color = html.find('meta', property='theme-color').get('content')
        return color
    else:
        return None


def get_page_phone_number(html: str, response: Any, index=0) -> str:
    """Get a phone number on a web page, the first found by default.
    Args:
        html (str): A body of HTML text.
        response (HTTPResponse): An HTTP response.
    Returns:
        (str): Returns the first phone number found.
    """
    try:
        phone = html.select('a[href*=callto]')[index].text
        return phone
    except:
        pass
    try:
        phone = re.findall(
            r'\(?\b[2-9][0-9]{2}\)?[-][2-9][0-9]{2}[-][0-9]{4}\b', response.text
        )[0]
        return phone
    except:
        pass
    try:
        phone = re.findall(
            r'\(?\b[2-9][0-9]{2}\)?[-. ]?[2-9][0-9]{2}[-. ]?[0-9]{4}\b', response.text
        )[-1]
        return phone
    except:
        print('Phone number not found')
        phone = ''
        return phone


def get_page_email(
        html: str,
        response: Any,
        index: Optional[int] = -1,
    ) -> str:
    """Get an email on a web page, the last email by default.
    Args:
        html (str): A body of HTML text.
        response (HTTPResponse): An HTTP response.
    Returns:
        (str): Returns the first email found on the page.
    """
    try:
        email = re.findall(
            r'([a-zA-Z0-9._-]+@[a-zA-Z0-9._-]+\.[a-zA-Z0-9_-]+)', response.text
        )[-1]
        return email
    except:
        pass
    try:
        email = html.select('a[href*=mailto]')[index].text
    except:
        print('Email not found')
        email = ''
        return email


def find_company_address():
    """
    TODO: Try to find a company's address from their website, then Google Maps.
    """
    raise NotImplementedError
    # street, city, state, zipcode = None, None, None, None
    # return street, city, state, zipcode


def find_company_linkedin():
    """
    TODO: Tru to find a company's LinkedIn URL. (Try to find LinkedIn on homepage?)
    """
    raise NotImplementedError


def find_company_url(company_name: str):
    """
    TODO: Find a company's website URL. (Google search for name?)
    """
    raise NotImplementedError


# === Google Drive Tools ===

def download_google_drive_file(drive_file, destination):
    """Download a public Google Drive file given its ID and a destination.
    Args:
        drive_file (str): A Google Drive ID or URL for a file.
        destination (str): The local file path and name.
    Credit: turdus-merula <https://stackoverflow.com/a/39225272/5021266>
    License: CC BY-SA 3.0 <https://creativecommons.org/licenses/by-sa/3.0/>
    """
    drive_id = drive_file
    if drive_id.startswith('https://drive.google'):
        drive_id = drive_id.split('/d/')[-1].split('/')[0]
    drive_base = 'https://docs.google.com/uc?export=download'
    drive_session = requests.Session()
    drive_response = drive_session.get(
        drive_base,
        params={'id': drive_id},
        stream=True,
    )
    drive_token = download_google_drive_file_confirm_token(drive_response)
    if drive_token:
        drive_response = drive_session.get(
            drive_base,
            params={'id': drive_id, 'confirm': drive_token},
            stream = True
        )
    download_google_drive_file_save_response(drive_response, destination)    


def download_google_drive_file_confirm_token(drive_response):
    """
    Credit: turdus-merula <https://stackoverflow.com/a/39225272/5021266>
    License: CC BY-SA 3.0 <https://creativecommons.org/licenses/by-sa/3.0/>
    """
    for k, v in drive_response.cookies.items():
        if k.startswith('download_warning'):
            return v
    return None


def download_google_drive_file_save_response(drive_response, destination):
    """
    Credit: turdus-merula <https://stackoverflow.com/a/39225272/5021266>
    License: CC BY-SA 3.0 <https://creativecommons.org/licenses/by-sa/3.0/>
    """
    CHUNK_SIZE = 32768
    with open(destination, 'wb') as f:
        for chunk in drive_response.iter_content(CHUNK_SIZE):
            if chunk:
                f.write(chunk)



"""
Analysis Model | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Analysis data model.
"""
# Standard imports
from dataclasses import dataclass
from typing import List

from .base import Model

@dataclass
class Analysis(Model):
    """Analyses represent scientific tests, such as cannabinoid
    analysis and pesticide screening. An analysis may contain
    multiple analytes, specific compound or substances that are
    being measured."""
    _collection = 'organizations/%s/analysis'
    analytes: list = List
    analyte_count: int = 0
    key: str = ''
    name: str = ''
    panel: bool = False
    price: float = 0.0
    public: bool = False



"""
Settings Models | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Settings data model.
"""
# Standard imports.
from dataclasses import dataclass

# Internal imports.
from .base import Model


@dataclass
class OrganizationSettings(Model):
    """An organizations's primary settings.
    An organization's `data_models` is the metadata that governs data
    collection. For example:

    "data_models": {
        "analytes": {
            "abbreviation": "AT",
            "current_count": "0",
            "fields": [
                {"key": "analyte_id", "label": "Analyte ID"},
                .
                .
                .
            ],
            "id_schema": "{{ abbreviation }}%y%m%d-{{ count }}",
            "singular": "analyte",
        },
        .
        .
        .
    }
    """
    _collection = 'organizations/%s/settings'
    traceability_provider: str = ''
    public: bool = False
    # data_models: dict = {}


@dataclass
class UserSettings(Model):
    """An organizations's primary settings."""
    _collection = 'users/%s/settings'
    public: bool = False



"""
Parse Sonoma Lab Works CoAs
Copyright (c) 2022 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
    Candace O'Sullivan-Sutherland <https://github.com/candy-o>
Created: 8/2/2022
Updated: 9/7/2022
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description:

    Parse Sonoma Lab Works CoA PDFs.

Data Points:

    ✓ analyses
    ✓ {analysis}_method
    ✓ {analysis}_status
    ✓ batch_number
    ✓ batch_size
    ✓ date_collected
    ✓ date_tested
    ✓ date_received
    ✓ distributor
    ✓ distributor_address
    ✓ distributor_street
    ✓ distributor_city
    ✓ distributor_county
    ✓ distributor_state
    ✓ distributor_zipcode
    ✓ distributor_latitude
    ✓ distributor_longitude
    - images
    x lab_results_url
    ✓ metrc_source_id
    ✓ producer
    ✓ producer_address
    ✓ producer_license_number
    ✓ producer_street
    ✓ producer_city
    ✓ producer_county
    ✓ producer_state
    ✓ producer_zipcode
    ✓ producer_latitude
    ✓ producer_longitude
    ✓ product_name
    ✓ product_type
    ✓ results
    ✓ sample_weight
    ✓ total_cannabinoids
    ✓ total_thc
    ✓ total_cbd
    ✓ total_terpenes
    ✓ sample_id (generated)
    - strain_name
    ✓ lab_id
    ✓ lab
    ✓ lab_email
    ✓ lab_phone
    ✓ lab_image_url
    ✓ lab_license_number
    ✓ lab_address
    ✓ lab_city
    ✓ lab_county (augmented)
    ✓ lab_state (augmented)
    ✓ lab_zipcode (augmented)
    ✓ lab_latitude (augmented)
    ✓ lab_longitude (augmented)

"""
# Standard imports.
from datetime import datetime
import json
from typing import Any, Optional

# External imports.
from dotenv import dotenv_values
import pandas as pd
import pdfplumber

# Internal imports.
from cannlytics.data.data import create_sample_id
from cannlytics.data.gis import search_for_address
from cannlytics.utils.constants import (
    ANALYSES,
    ANALYTES,
    STANDARD_FIELDS,
)
from cannlytics.utils.utils import (
    convert_to_numeric,
    snake_case,
)

# It is assumed that the lab has the following details.
SONOMA =  {
    'coa_algorithm': 'sonoma.py',
    'coa_algorithm_entry_point': 'parse_sonoma_coa',
    'lims': 'Sonoma Lab Works',
    'lab': 'Sonoma Lab Works',
    'lab_license_number': 'C8-0000015-LIC',
    'lab_image_url': 'https://images.squarespace-cdn.com/content/v1/6072f29fa4b8d3121b2f9fbc/1623325374642-KXUXXMCSCVBYPYMFPTUS/slw-logo.png?format=1500w',
    'lab_address': '1201 Corporate Center Parkway, Santa Rosa, CA 95407',
    'lab_street': '1201 Corporate Center Parkway',
    'lab_city': 'Santa Rosa',
    'lab_county': 'Sonoma',
    'lab_state': 'CA',
    'lab_zipcode': '95407',
    'lab_latitude': 38.421600,
    'lab_longitude': -122.752990,
    'lab_phone': '707-757-7757',
    'lab_email': 'testing@sonomalabworks.com',
    'lab_website': 'https://www.sonomalabworks.com/',
}


def parse_sonoma_coa(
        parser,
        doc: Any,
        google_maps_api_key: Optional[str] = None,
        **kwargs,
    ) -> Any:
    """Parse a Sonoma Lab Works CoA PDF.
    Args:
        doc (str or PDF): A PDF file path or pdfplumber PDF.
    Returns:
        (dict): The sample data.
    """
    # Read the PDF.
    obs = {}
    if isinstance(doc, str):
        report = pdfplumber.open(doc)
        obs['coa_pdf'] = doc.split('/')[-1]
    else:
        report = doc
        obs['coa_pdf'] = report.stream.name.split('/')[-1]

    # Define fields to collect.
    analysis = None
    analyses = []
    columns = []
    results = []
    units = None

    # Get all of the tables.
    tables = []
    for page in report.pages:
        tables.extend(page.extract_tables())

    # Get the details from the front page.
    tables = report.pages[0].extract_tables()
    for table in tables:
        for row in table:
            initial_value = row[0]

            # Skip blank rows.
            if not initial_value:
                continue

            # Get standard fields.
            if ':' in initial_value:
                if len(row) == 2:
                    key = snake_case(initial_value)
                    key = STANDARD_FIELDS.get(key, key)
                    value = row[1]
                    obs[key] = value
                elif len(row) == 4:
                    key = snake_case(initial_value)
                    key = STANDARD_FIELDS.get(key, key)
                    value = row[1]
                    second_key = snake_case(row[2])
                    second_key = STANDARD_FIELDS.get(second_key, second_key)
                    second_value = row[3]
                    obs[key] = value
                    obs[second_key] = second_value

            # Get totals.
            elif 'Total' in initial_value:
                key = STANDARD_FIELDS.get(initial_value, snake_case(initial_value))
                obs[key] = convert_to_numeric(row[1])

            # Get analysis statuses and minor analysis results.
            elif 'Pass' in str(row[-1]) or 'Fail' in str(row[-1]):
                name = initial_value.split('(')[0].strip()
                analysis = ANALYSES.get(name, snake_case(name))
                status = row[-1].lower()
                obs[f'{analysis}_status'] = status

                # Hot-fix: Also collect minor analyses.
                special_analyses = {
                    'moisture_content': {'analysis': 'moisture_content', 'units': 'percent', 'limit': 15},
                    'water_activity': {'analysis': 'water_activity', 'units': 'percent', 'limit': 0.65},
                    'foreign_matter': {'analysis': 'foreign_matter', 'units': 'percent', 'limit': None},
                }
                if analysis in special_analyses.keys():
                    special_analysis = special_analyses[analysis]
                    results.append({
                        'key': special_analysis['analysis'],
                        'name': name,
                        'value': convert_to_numeric(row[1]),
                        'units': special_analysis['units'],
                        'limit': special_analysis.get('limit'),
                        'status': status,
                    })

    # Get the analyses and results.
    tables = report.pages[1].extract_tables()
    for table in tables:
        for row in table:

            # Skip the detail rows.
            if len(row) == 2 or not row[0]:
                continue

            # Find the analysis.
            elif '\n' in row[0]:
                parts = row[0].split('\n')
                analysis = ANALYSES.get(parts[0], snake_case(parts[0]))
                analyses.append(analysis)
                for part in parts:
                    if 'Method:' in part:
                        method = part \
                            .split('Method:')[-1] \
                            .split('Date Tested:')[0] \
                            .strip()
                        obs[f'{analysis}_method'] = method
                        break

            # Find the table columns.
            elif row[0] == 'Target Analyte' or row[0] == 'Microbiological Assay':
                columns = [x for x in row if x is not None]

                # Identify the units.
                units = None
                for column in columns:
                    if column and '(' in column:
                        units = column.split('(')[-1].split(')')[0]

            # Parse the result, standardizing the analyte keys and fields.
            else:
                values = [x for x in row if x is not None]
                result = {'analysis': analysis, 'units': units}
                recorded_keys = []
                for i, column in enumerate(columns):
                    if not column:
                        continue
                    elif i == 0:
                        analyte = snake_case(values[i])
                        result['key'] = ANALYTES.get(analyte, analyte)
                    key = STANDARD_FIELDS.get(column, snake_case(column))
                    if key in recorded_keys:
                        continue
                    try:
                        result[key] = convert_to_numeric(values[i])
                    except IndexError:
                        result[key] = None
                    recorded_keys.append(key)
                results.append(result)

    # Hot-fix for cannabinoid units.
    for i, result in enumerate(results):
        results[i]['units'] = 'percent'

    # Hot-fix for sample type and sample matrix.
    obs['product_type'] = ', '.join([obs['matrix'], obs['product_type']])

    # Hot-fix for `date_tested`, e.g. '7/6/2021 - 7/8/2021'.
    obs['date_tested'] = obs['date_tested'].split(' - ')[-1]

    # Turn dates to ISO format.
    date_columns = [x for x in obs.keys() if x.startswith('date')]
    for date_column in date_columns:
        try:
            obs[date_column] = pd.to_datetime(obs[date_column]).isoformat()
        except:
            pass

    # Try to get the producer's address.
    try:
        location = search_for_address(obs['producer_address'], api_key=google_maps_api_key)
        for key, value in location.items():
            obs[f'producer_{key}'] = value
    except:
        pass

    # Try to get the distributor's address.
    try:
        location = search_for_address(obs['distributor_address'], api_key=google_maps_api_key)
        for key, value in location.items():
            obs[f'distributor_{key}'] = value
    except:
        pass

    # Future work: Standardize `product_type`.

    # Aggregate results.
    obs['analyses'] = list(set(analyses))
    obs['results'] = results

    # Finish data collection with a freshly minted sample ID.
    obs['sample_id'] = create_sample_id(
        private_key=json.dumps(results),
        public_key=obs['product_name'],
        salt=obs['producer'],
    )
    obs['coa_parsed_at'] = datetime.now().isoformat()
    return {**SONOMA, **obs}


# === Test ===
if __name__ == '__main__':

    from cannlytics.data.coas import CoADoc

    # Get Google Maps API Key.
    try:
        config = dotenv_values('../../../.env')
        google_maps_api_key = config['GOOGLE_MAPS_API_KEY']
    except:
        google_maps_api_key = None

    # [✓] TEST: Parse a Sonoma Lab Works CoA PDF.
    parser = CoADoc()
    doc = '../../../tests/assets/coas/LemonTree.pdf'
    lab = parser.identify_lims(doc)
    assert lab == 'Sonoma Lab Works'
    data = parse_sonoma_coa(parser, doc, google_maps_api_key=google_maps_api_key)
    assert data is not None



"""
Parse Veda Scientific CoAs
Copyright (c) 2022 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
    Candace O'Sullivan-Sutherland <https://github.com/candy-o>
Created: 7/15/2022
Updated: 9/7/2022
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description:

    Parse a Veda Scientific CoA PDF.

Data Points:

    ✓ analyses
    - {analysis}_method
    ✓ analysis_type
    ✓ {analysis}_status
    x coa_urls
    ✓ date_tested
    - date_received
    ✓ distributor
    ✓ distributor_address
    ✓ distributor_street
    ✓ distributor_city
    ✓ distributor_state
    ✓ distributor_zipcode
    ✓ distributor_license_number
    x images
    x lab_results_url
    ✓ producer
    ✓ producer_address
    ✓ producer_street
    ✓ producer_city
    ✓ producer_state
    ✓ producer_zipcode
    ✓ producer_license_number
    ✓ product_name
    ✓ lab_id
    ✓ project_id
    ✓ product_type
    ✓ batch_number
    ✓ metrc_lab_id
    ✓ metrc_source_id
    ✓ date_collected
    ✓ date_received
    ✓ sample_size
    ✓ serving_size
    ✓ servings_per_package
    ✓ sample_weight
    - results
    ✓ status
    - total_cannabinoids (calculated)
    - total_thc
    - total_cbd
    - total_terpenes (calculated)
    ✓ sample_id (generated)
    ✓ lab
    ✓ lab_image_url
    ✓ lab_license_number
    ✓ lab_address
    ✓ lab_street
    ✓ lab_city
    ✓ lab_county (augmented)
    ✓ lab_state
    ✓ lab_zipcode
    ✓ lab_phone
    ✓ lab_email
    ✓ lab_website
    ✓ lab_latitude (augmented)
    ✓ lab_longitude (augmented)

FIXME:

    This algorithm is under development!
    Please email dev@cannlytics.com if you want to help out.

    - [ ] Add `results`.

"""
# Standard imports.
from ast import literal_eval
from datetime import datetime
import json
import re
from typing import Any, Optional

# External imports.
import pandas as pd
import pdfplumber

# Internal imports.
from cannlytics.data.data import create_sample_id, find_first_value
# from cannlytics.utils.constants import ANALYSES, ANALYTES
from cannlytics.utils.utils import (
    convert_to_numeric,
    snake_case,
    split_list,
    strip_whitespace,
)


# It is assumed that the lab has the following details.
VEDA_SCIENTIFIC = {
    'coa_algorithm': 'veda.py',
    'coa_algorithm_entry_point': 'parse_veda_coa',
    'url': 'vedascientific.co',
    'lims': 'Veda Scientific',
    'lab': 'Veda Scientific',
    'lab_image_url': 'https://images.squarespace-cdn.com/content/v1/5fab1470f012f739139935ac/58792970-f502-4e1a-ac29-ddca27b43266/Veda_Logo_Horizontal_RGB_Large.png?format=1500w', # <- Get this data.
    'lab_license_number': '', # <-- Get this static data point.
    'lab_address': '1601 W Central Ave Building A Unit A, Lompoc, CA',
    'lab_street': '1601 W Central Ave Building A Unit A',
    'lab_city': 'Lompoc',
    'lab_county': 'Santa Barbara',
    'lab_state': 'CA',
    'lab_zipcode': '93436',
    'lab_phone': '(805) 324-7728',
    'lab_email': 'info@vedascientific.co',
    'lab_website': 'vedascientific.co',
    'lab_latitude': 34.661520,
    'lab_longitude': -120.476520,
}

# It is assumed that there are the following analyses on each CoA.
VEDA_SCIENTIFIC_ANALYSES = {
    'cannabinoids': {
        'name': 'Cannabinoids',
        'columns': ['name', 'lod', 'loq', 'mg_g', 'value',
                    'mg_per_serving', 'mg_per_package'],
        'units': 'mg/g',
        # FIXME: It's better to have analytes dynamic.
        # 'analytes': [
        #     'cbd',
        #     'cbda',
        #     'cbdv',
        #     'cbdva',
        #     'cbg',
        #     'cbga',
        #     'cbl',
        #     'cbc',
        #     'cbca',
        #     'cbn',
        #     'cbna',
        #     'delta_9_thc',
        #     'thca',
        #     'delta_8_thc',
        #     'thcv',
        #     'thcva',
        # ],
    },
    'terpenes': {
        'columns': ['name', 'lod', 'loq', 'mg_g', 'value',
                    'mg_per_serving', 'mg_per_package'],
        'units': 'mg/g',
    },
    'moisture': {
        'columns': ['name', 'value'],
        'units': 'percent',
    },
    'water_activity': {
        'columns': ['name', 'value', 'limit', 'status'],
        'units': 'aW',
    },
    'pesticides': {
        'columns': ['name', 'lod', 'loq', 'limit', 'value', 'status'],
        'units': 'ug/g',
    },
    'mycotoxins': {
        'columns': ['name', 'lod', 'loq', 'limit', 'value', 'status'],
        'units': 'ug/kg',
    },
    'heavy_metals': {
        'columns': ['name', 'lod', 'loq', 'limit', 'value', 'status'],
        'units': 'ug/g',
    },
    'foreign_matter': {
        'analytes': [
            'mold',
        ],
        'columns': ['name', 'limit', 'status'],
        'units': 'percent',
    },
    'microbials': {
        'columns': ['name', 'value', 'status'],
        'units': 'count',
    },
    'residual_solvents': {
        'columns': ['name', 'lod', 'loq', 'limit', 'value', 'status'],
        'units': 'ug/g',
    },
}

# It is assumed that the CoA has the following parameters.
VEDA_SCIENTIFIC_COA = {
    'coa_qr_code_index': None,
    'coa_image_index': 0,
    'coa_page_area': [0, 0.095, 1, 0.85],
    'coa_distributor_area': [0, 0.1, 0.5, 0.215],
    'coa_producer_area': [0.5, 0.1, 1, 0.215],
    'coa_sample_details_area': [0, 0.215, 1, 0.375],
    'coa_analyses': {
        'CANNABINOID': 'cannabinoids',
        'TERPENES': 'terpenes',
        'MOISTURE': 'moisture',
        'WATER ACTIVITY': 'water_activity',
        'PESTICIDES': 'pesticides',
        'MYCOTOXIN': 'mycotoxins',
        'FOREIGN MATERIALS': 'foreign_matter',
        'RESIDUAL SOLVENTS': 'residual_solvents',
        'HEAVY METALS': 'heavy_metals',
        'MICROBIAL IMPURITIES': 'microbes',
    },
    'coa_fields': {
        'Sample Name': 'product_name',
        'Collected': 'date_collected',
        'Lab ID': 'lab_id',
        'ReceiveEd': 'date_received',
        'Order Number': 'order_number',
        'Client Batch Size': 'batch_size',
        'Matrix Type': 'product_type',
        'Sample Size': 'sample_size',
        'Client Batch #': 'batch_number',
        'Serving Size': 'serving_size',
        'Lab Metrc Id': 'metrc_lab_id',
        'Servings': 'servings_per_package',
        'Client Metrc Id': 'metrc_source_id',
        'Density': 'sample_weight',
    },
    'coa_skip_values': [],
    'coa_replacements': [
        {'text': ' Per package', 'key': ''},
        {'text': 'Potency', 'key': 'Cannabinoids'},
        {'text': 'Foreign Materials', 'key': 'Foreign Matter'},
    ],
}


def get_page_rows(
        pdf_file: Any,
        index: list,
        analytes: dict,
    ) -> list:
    """Get the rows a given page.
    Split long rows at the first analyte!
    Args:

    Returns:
    """
    analyte_keys = list(set(analytes.keys()))
    analyte_values = list(set(analytes.values()))
    page_rows, texts = [], []
    for i in index:
        page = pdf_file.pages[i]
        text = page.extract_text()
        text = text \
            .replace('\xa0\xa0', '\n') \
            .replace('\xa0', ',') \
            .split('\n')
        texts.append(text)
    
    # TODO: Split into separate function?
    for text in texts:
        for row in text:
            parts = row.split(' ')
            initial_value = parts[0]
            analyte_key =  snake_case(strip_whitespace(initial_value))
            if initial_value in analyte_values:
                analyte_key = analytes[analyte_key]
            if analyte_key in analyte_keys:
                second_index = None
                for i, part in enumerate(parts[1:]):
                    second_key = snake_case(strip_whitespace(part))
                    if second_key in analyte_keys:
                        second_index = i + 1
                        break
                if second_index is None:
                    page_rows.append(parts)
                else:
                    double_rows = split_list(parts, second_index)
                    page_rows.append(double_rows[0])
                    page_rows.append(double_rows[1])
            else:
                page_rows.append(parts)
    return page_rows


def parse_veda_coa(parser, doc: Any, **kwargs) -> dict:
    """Parse a Veda Scientific CoA PDF.
    Args:
        doc (str or PDF): A PDF file path or pdfplumber PDF.
    Returns:
        (dict): The sample data.
    """
    # Read the PDF.
    obs = {}
    if isinstance(doc, str):
        report = pdfplumber.open(doc)
        obs['coa_pdf'] = doc.replace('\\', '/').split('/')[-1]
    else:
        report = doc
        obs['coa_pdf'] = report.stream.name.replace('\\', '/').split('/')[-1]
    front_page = report.pages[0]
    w, h = front_page.width, front_page.height

    # Get the lab-specific CoA page areas.
    coa_parameters = VEDA_SCIENTIFIC_COA

    # If needed: Get lab CoA specific fields.
    coa_fields = coa_parameters['coa_fields']
    coa_replacements = coa_parameters['coa_replacements']
    sample_details_fields = list(coa_fields.keys())
    standard_analyses = coa_parameters['coa_analyses']
    # skip_values = coa_parameters['coa_skip_values']

    # Optional: Get the image data.
    # image_index = coa_parameters['coa_image_index']
    # obs['image_data'] = self.get_pdf_image_data(report.pages[0], image_index)
    obs['images'] = []

    # Get all distributor details.
    x0, y0, x1, y1 = tuple(coa_parameters['coa_distributor_area'])
    distributor_area = (x0 * w, y0 * h, x1 * w, y1 * h)
    crop = front_page.within_bbox(distributor_area)
    text = crop.extract_text()
    block = text.split('\n')
    distributor = text \
        .split('License Number')[0] \
        .split('Business Name')[-1] \
        .replace('\n', ' ').strip()
    license_number = text \
        .split('License Number')[-1] \
        .split('Address')[0].strip()
    street = block[-2]
    locales = block[-1].split(',')
    city = locales[0]
    subparts = locales[-1].strip().split(' ')
    state, zipcode = subparts[-2], subparts[-1]
    address = ', '.join([street, block[-1]])
    obs['distributor'] = distributor
    obs['distributor_address'] = address
    obs['distributor_street'] = street
    obs['distributor_city'] = city
    obs['distributor_state'] = state
    obs['distributor_zipcode'] = zipcode
    obs['distributor_license_number'] = license_number

    # Get all producer details.
    x0, y0, x1, y1 = tuple(coa_parameters['coa_producer_area'])
    producer_area = (x0 * w, y0 * h, x1 * w, y1 * h)
    crop = front_page.within_bbox(producer_area)
    text = crop.extract_text()
    block = text.split('\n')
    producer = text \
        .split('License Number')[0] \
        .split('Business Name')[-1] \
        .replace('\n', ' ').strip()
    license_number = text \
        .split('License Number')[-1] \
        .split('Address')[0].strip()
    street = block[-2]
    locales = block[-1].split(',')
    city = locales[0]
    subparts = locales[-1].strip().split(' ')
    state, zipcode = subparts[-2], subparts[-1]
    address = ', '.join([street, block[-1]])
    obs['producer'] = producer
    obs['producer_address'] = address
    obs['producer_street'] = street
    obs['producer_city'] = city
    obs['producer_state'] = state
    obs['producer_zipcode'] = zipcode
    obs['producer_license_number'] = license_number

    # Get the sample details.
    x0, y0, x1, y1 = tuple(coa_parameters['coa_sample_details_area'])
    sample_details_area = (x0 * w, y0 * h, x1 * w, y1 * h)
    crop = front_page.within_bbox(sample_details_area)
    text = crop.extract_text()
    for r in coa_replacements:
        text = text.replace(r['text'], r['key'])
    details = re.split('\n|' + '|'.join(sample_details_fields), text)
    index = 0
    for i, detail in enumerate(details[1:]):
        if detail:
            field = sample_details_fields[index]
            key = coa_fields[field]
            obs[key] = detail.lstrip(':').strip()
            index += 1 

    # Get the front page tables.
    tables = front_page.extract_tables()

    # Get the overall status.
    table_data = tables[0][0][0]
    analysis_details = table_data.replace('\xa0\xa0\xa0\xa0', '\n').split('\n')
    mm, dd, yyyy = analysis_details[2].split(':')[-1].strip().split('/')
    date_tested = f'{yyyy}-{mm}-{dd}'
    obs['date_tested'] = date_tested
    obs['analysis_type'] = analysis_details[0]
    obs['status'] = analysis_details[-1].split(':')[-1].strip().lower()

    # Get the statuses from the front page.
    analyses = []
    x0, y0, x1, y1 = tuple(coa_parameters['coa_page_area'])
    page_area = (x0 * w, y0 * h, x1 * w, y1 * h)
    crop = front_page.within_bbox(page_area)
    text = crop.extract_text()
    summary = text.split('Sample Certification')[0]
    summary = summary.split('SAFETY SUMMARY')[-1]
    summary = summary.replace('\xa0', '\n').split('\n')
    for row in summary:
        if row:
            parts = row.strip().split(' ')
            analysis_name = ' '.join(parts[:len(parts) - 1])
            for r in coa_replacements:
                analysis_name = analysis_name.replace(r['text'], r['key'])
            analysis = snake_case(analysis_name)
            status = parts[-1].strip().lower()
            obs[f'{analysis}_method'] = status
            # Future work: Ensure this excludes not tested analyses.
            # Will need to see an example CoA with not tested analyses.
            if status != 'not tested':
                analyses.append(analysis)
    
    # TODO: Get cannabinoid totals.
    # block = text.split('\n')
    # # - total_thc
    # # - total_cbd
    # # - total_cannabinoids

    # Get the lab license number.
    table_data = tables[-1][-1][0]
    lab_details = table_data.replace('\xa0', '\n').split('\n')
    lab_details = [x.strip() for x in lab_details if x.replace('|', '').strip() != '' ]
    for detail in lab_details:
        if 'license' in detail.lower():
            obs['lab_license_number'] = detail.split(':')[-1].strip()

    # Get all of the `results` rows.
    all_lines = []
    for page in report.pages[1:]:
        crop = page.within_bbox(page_area)
        rows = parser.get_page_rows(crop)
        for row in rows:
            if row in all_lines:
                pass
            else:
                all_lines.append(row)

    # Brute-force: The following probably needs to be refactored.
    analysis_keys = list(standard_analyses.keys())

    # Get all of the `results`!
    current_analysis = []
    collected_analytes = []
    results = []
    for line in all_lines:

        # Skip nuisance lines.
        if len(line) == 1 or ':' in line:
            continue

        # Get the analysis of the table, handling two columns.
        elif 'RESULTS' in line:
            current_analysis = []
            for key in analysis_keys:
                if key in line:
                    current_analysis.append(standard_analyses[key])
        
        # Get the analysis methods.
        elif 'Method:' in line:
            parts = line.split('Method:')
            methods = [
                x.split(' ,')[0].split('Procedures:')[0].strip() for x in parts if x
            ]
            for i, analysis in enumerate(current_analysis):
                try:
                    obs[f'{analysis}_method'] = methods[i]
                except IndexError:
                    pass

        # Optional: Identify the columns?

        # FIXME: Collect each individual result.
        else:

            # Add keys / values to the result, using standard fields.
            first_value = find_first_value(line)
            name = line[:first_value].strip()
            analyte = snake_case(name)
            analyte = parser.fields.get(analyte, analyte)
            values = line[first_value:].strip().split(' ')
            values = [x for x in values if x]

            # Create a result object.
            result = {
                # 'analysis': analysis, # FIXME: Add `analysis.`
                'key': analyte,
                'name': name,
                # 'units': units, # FIXME: Add `units.
            }

            # TODO: Get the columns.
            
            # Match the values to the columns.
            # try:
            #     for i, v in enumerate(values):
            #         # FIXME: Get the correct column field here.
            #         key = parser.fields[i + 1]
            #         value = convert_to_numeric(v)
            #         result[key] = value
            # except IndexError:
            #     continue

            # Record the result.
            results.append(result)

    # Turn dates to ISO format.
    date_columns = [x for x in obs.keys() if x.startswith('date')]
    for date_column in date_columns:
        try:
            obs[date_column] = pd.to_datetime(obs[date_column]).isoformat()
        except:
            pass

    # Return the sample with a freshly minted sample ID.
    obs['analyses'] = list(set(analyses))
    obs['date_tested'] = date_tested
    obs['results'] = results
    obs['sample_id'] = create_sample_id(
        private_key=json.dumps(results),
        public_key=obs['product_name'],
        salt=producer,
    )
    return {**VEDA_SCIENTIFIC, **obs}


# === Tests ===
if __name__ == '__main__':

    from cannlytics.data.coas import CoADoc

    # Specify where your test CoA lives.
    DATA_DIR = './tests/assets/coas/veda-scientific'
    doc = f'{DATA_DIR}/Veda Scientific Sample COA.pdf'

    # [✓] TEST: Detect the lab / LIMS that generated the CoA.
    parser = CoADoc()
    known_lims = parser.identify_lims(doc)
    assert known_lims == 'Veda Scientific'

    # [ ] TEST: Parse the CoA.
    parser = CoADoc()
    data = parse_veda_coa(parser, doc)
    assert data is not None



"""
CoADoc | Get MCR Labs Test Result Data
Copyright (c) 2022 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
    Candace O'Sullivan-Sutherland <https://github.com/candy-o>
Created: 7/13/2022
Updated: 2/17/2023
License: MIT License <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description:

    Tools to collect MCR Labs' publicly published lab results.
    MCR Labs' COA PDFs are generally laid out as follows:

        Page 1 Lab/Product/Test information
        Page 2 Cannabinoid Test result
        Page 3 Microbe Test results
        Page 4 Mycotoxin and Heavy Metal Test results
        Page 5 Pesticide Test results
        Page 6 Terpenes Test results
        Pages 7 and 8 are for QC/QA and are not used.

Data Points:

    ✓ analyses
    ✓ {analysis}_method
    ✓ date_tested
    ✓ images
    ✓ lab
    ✓ lab_website
    ✓ lab_results_url
    ✓ product_name
    ✓ product_type
    ✓ producer
    ✓ results
        ✓ analysis
        ✓ key
        ✓ name
        ✓ units
        ✓ value
        ✓ lod
        ✓ loq
    ✓ sample_id (generated)
    ✓ total_cannabinoids
    ✓ total_terpenes

Data Sources:
    
    - MCR Labs Test Results
    URL: <https://reports.mcrlabs.com>

Future development:

    - [ ] Implement a function to get all of a given client's lab results.
    - Optional: Create necessary data dirs automatically.
    - Optional: Function to download any pre-existing results.

"""
# Standard imports.
from datetime import datetime
import json
import math
import re
from time import sleep
from typing import Any, Optional

# External imports.
from bs4 import BeautifulSoup
import pandas as pd
import pdfplumber
import requests

# Internal imports.
from cannlytics import __version__
from cannlytics.data.data import create_hash, create_sample_id
from cannlytics.utils.constants import (
    ANALYSES,
    ANALYTES,
    DECARB,
    DEFAULT_HEADERS,
    STANDARD_UNITS,
)
from cannlytics.utils.utils import (
    convert_to_numeric,
    format_iso_date,
    snake_case,
    strip_whitespace,
)
from datasets import load_dataset

# Suppress `datasets` warning:
# Deprecated argument(s) used in 'dataset_info': token.
# Will not be supported from version '0.12'.
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)


# It is assumed that the lab has the following details.
MCR_LABS = {
    'coa_algorithm': 'mcrlabs.py',
    'coa_algorithm_entry_point': 'parse_mcrlabs_coa',
    'lims': 'MCR Labs',
    'url': 'https://reports.mcrlabs.com',
    'lab': 'MCR Labs',
    'lab_website': 'https://mcrlabs.com',
    'lab_license_number': 'IL281277',
    'lab_image_url': '',
    'lab_address': '85 Speen St. Lower Level, Framingham, MA 01701',
    'lab_street': '85 Speen St. Lower Level',
    'lab_city': 'Framingham',
    'lab_county': 'Middlesex',
    'lab_state': 'MA',
    'lab_zipcode': '01701',
    'lab_latitude': 42.310820,
    'lab_longitude': -71.386920,
    'lab_phone': '508-872-6666',
    'lab_email': 'hello@mcrlabs.com',
}

# It is assumed that there are the following analyses on each COA.
MCR_LABS_COA = {
    'fields': {
        'Sample ID #': 'lab_id',
        'Sample Name': 'product_name',
        'Batch': 'batch_number',
        'Matrix': 'product_type',
        'Date Received' : 'date_received',
        'Date Tested': 'date_tested',
        'Sample Weight': 'sample_weight',
        'Serving size weight': 'serving_size',
        'Analyte': 'name',
        'Cannabinoid': 'skip',
        'Conc.': 'value',
        'Test ID': 'skip',
        'Test Analysis': 'name',
        'Result': 'value',
        'LOD': 'lod',
        'LOQ': 'loq',
        'Limits': 'limit',
        'Disposition': 'status',
        'Results': 'value',
        'Unit': 'skip',
    }
}


def get_mcr_labs_sample_count(
        per_page: Optional[int] = 30,
        headers: Optional[Any] = None,
    ) -> dict:
    """Get the number of samples and pages of MCR Labs samples.
    Args:
        per_page (int): The number of results displayed per page.
        headers (dict): Headers for the HTTP request (optional).
    Returns:
        (dict): A dictionary with `count` and `pages` keys.
    """
    base = MCR_LABS['url']
    url = f'{base}/products-weve-tested'
    if headers is None:
        headers = DEFAULT_HEADERS
    response = requests.get(url, headers=headers)
    soup = BeautifulSoup(response.content, 'html.parser')
    element = soup.find('span', attrs={'id': 'found_posts'})
    count = int(element.text.replace(',', ''))
    pages = math.ceil(count / per_page)
    return {'count': count, 'pages': pages}


def get_mcr_labs_sample_details(
        parser,
        lab_id: str,
        headers: Optional[Any] = None,
        standard_analyses: Optional[Any] = None,
        standard_analytes: Optional[Any] = None,
        **kwargs,
    ) -> dict:
    """Get the details for a specific MCR Labs test sample.
    Args:
        parser (CoADoc): A CoADoc client for standardization.
        lab_id (str): A sample ID number or the `lab_results_url`.
        headers (dict): Headers for the HTTP request (optional).
        standard_analyses (dict): An optional mapping of lab-specific
            analyses to standard analyses.
        standard_analytes (dict): An optional mapping of lab-specific
            analyses to standard analytes.
    Returns:
        (dict): A dictionary of sample details.
    """
    # Get the sample page.
    obs = {}
    if lab_id.startswith('https'):
        url = lab_id
    else:
        base = MCR_LABS['url']
        url = f'{base}/reports/{lab_id}'
    if headers is None:
        headers = DEFAULT_HEADERS
    response = requests.get(url, headers=headers)
    soup = BeautifulSoup(response.content, 'html.parser')

    # Get the lab.
    try:
        element = soup.find('div', attrs={'class': 'rd_date'})
        obs['lab'] = strip_whitespace(element.text).split('by ')[-1]
    except:
        print('Failed to find lab:', lab_id)
        obs['lab'] = ''
    
    # Get a map of lab-specific analyses, analytes, and fields to the standard.
    if standard_analyses is None:
        standard_analyses = ANALYSES
    if standard_analytes is None:
        standard_analytes = ANALYTES
    
    # Get the sample image, if not already collected.
    if not obs.get('images'):
        attrs = {'class': 'report_image'}
        image_url = soup.find('img', attrs=attrs)['src']
        filename = image_url.split('/')[-1]
        obs['images'] = [{'url': image_url, 'filename': filename}]

    # Get the date tested, if not already collected.
    if not obs.get('date_tested'):
        text = soup.find('div', attrs={'class': 'rd_date'}).text
        date = text.split('Tested ')[-1].split(' for')[0]
        obs['date_tested'] = format_iso_date(date)

    # Get the product name and producer, if not already collected.
    if not obs.get('product_name'):
        details = soup.find('div', attrs={'id': 'client'})
        product_name = details.find('b').text
        obs['product_name'] = product_name
        obs['producer'] = details.find('h4').text.replace(product_name, '')

    # Get the product type, if not already collected.
    # Optional: Standardize the product types?
    if not obs.get('product_type'):
        el = soup.find('div', attrs={'class': 'abb'})
        obs['product_type'] = el.attrs['class'][-1].replace('repor_', '')

    # Get the total cannabinoids and total terpenes, if not already collected.
    if not obs.get('total_cannabinoids'):
        details = soup.find('div', attrs={'class': 'client_detail'})
        values = details.find_all('h3')
        keys = details.find_all('b')
        for i, el in enumerate(values):
            key = snake_case(strip_whitespace(keys[i].text))
            value = convert_to_numeric(strip_whitespace(el.text), strip=True)
            obs[key] = value

    # Optional: Get any product serving size (for edibles?).

    # Get the analysis methods.
    details = soup.find_all('div', attrs={'class': 'calc-expl'})
    for detail in details:
        pars = detail.find_all('p')
        for par in pars:
            if 'quantified' in par.text:
                text = strip_whitespace(par.text)
                analysis = snake_case(text.split(' are quantified')[0])
                analysis = standard_analyses.get(analysis, analysis)
                method = text.split('quantified using ')[-1]\
                        .replace('.', '').title()
                obs[f'{analysis}_method'] = method
    
    # Get the sample test results.
    analyses, results = [], []

    # Get the cannabinoids.
    scripts = soup.find_all('script')
    for script in scripts:
        if script.string:
            if 'dataProvided' in script.string:
                break
    block = script.string.split('const dataProvided = ')[-1].split(';')[0]
    block = strip_whitespace(block).replace(' ', '').replace(',]', ']')
    cannabinoids = json.loads(block)
    analyses.append('cannabinoids')

    # Determine the cannabinoids units.
    units = None
    try:
        assert '%' in soup.find('div', attrs={'class': 'rd_can_table'}).text
        units = 'percent'
    except AttributeError:
        table = soup.find('table', attrs={'class': 'safetytable'})
        thead = table.find('thead')
        units = snake_case(thead.find_all('th', limit=2)[-1].text)

    # Record the cannabinoids.
    for analyte in cannabinoids:
        if analyte['label'] == 'TotalCannabinoids':
            continue
        key = snake_case(analyte['key'].replace('-', ''))
        key = standard_analytes.get(key, key)
        results.append({
            'analysis': 'cannabinoids',
            'key': key,
            'name': analyte['label'],
            'value': analyte['perc'],
            'units': units,
        })

    # Get the results for all other analyses.
    values = ['name', 'value', 'lod' ,'loq']
    tables = soup.find_all('table', attrs={'class': 'safetytable'})
    for table in tables:

        # Get the analysis of the table.
        el = table.find_previous('h4')
        try:
            assert el.parent['id'] == 'client'
            analysis = 'cannabinoids'
        except KeyError:
            text = strip_whitespace(el.text)
            text = text \
                .replace('Screen', '') \
                .replace('Contaminant', '') \
                .replace('Profile', '')
            analysis = snake_case(text.strip())
            analysis = standard_analyses.get(analysis, analysis)
        analyses.append(analysis)

        # Optional: Break up aflatoxins and ochratoxin A.
        # Aflatoxin (B1, B2, G1, G2) &amp; Ochratoxin A

        # Get each result field.
        current_units = None
        rows = table.find('tbody').find_all('tr')
        for row in rows:
            result = {'analysis': analysis}

            # Precisely extract results if possible.
            try:
                name = row.find('td', attrs={'class': 'sr_label'}).text.strip()
            except AttributeError:
                name = None
            try:
                value = row.find('td', attrs={'class': 'sr_result'}).text.strip()
            except:
                value = None
            units = None
            try:
                lod = row.find('td', attrs={'class': 'sr_limit'}).text
                loq = row.find('td', attrs={'class': 'sr_limit'}).text
                result['lod'] = convert_to_numeric(lod, strip=True)
                result['loq'] = convert_to_numeric(loq, strip=True)
                result['units'] = re.sub('[\d\.]', '', lod).strip()
            except AttributeError:
                pass
            if name and value:
                analyte = snake_case(name)
                analyte = standard_analytes.get(analyte, analyte)
                result['key'] = analyte
                result['value'] = convert_to_numeric(value)
                results.append(result)

            # Otherwise, parse results to the best of our abilities.
            if name is None and value is None:
                cells = row.find_all('td')
                if len(cells) > 1:
                    for i, cell in enumerate(cells):
                        value, units = None, None
                        try:
                            key = values[i]
                        except IndexError:
                            key = f'limit_{i}'
                        if key == 'name':
                            analyte = snake_case(cell.text)
                            analyte = standard_analytes.get(analyte, analyte)
                            result['key'] = analyte
                        elif key == 'value':
                            value = cell.text
                            units = re.sub('[\d\.]', '', value) \
                                .replace('%', 'percent') \
                                .replace('<', '') \
                                .strip()
                            if not units or units == 'ND' or units == 'Not Detected':
                                limit = result.get('lod', result.get('loq', result.get('limit', '')))
                                units = re.sub('[\d\.]', '', limit).replace('%', 'percent').strip()
                            if units != 'Negative' and units != 'Positive':
                                value = convert_to_numeric(re.sub('[^\d\.]', '', value))
                                result['value'] = value
                                result['units'] = units
                                current_units = units
                            else:
                                result['value'] = units
                                result['units'] = current_units
                        else:
                            result[key] = convert_to_numeric(cell.text, strip=True)

                    # Record the result, if present.
                    if not value and not units:
                        continue
                    results.append(result)

                # Get all of the non-detects, using the current units.
                # Note: This is crudely using units from the prior table.
                elif row.has_attr('class') and row.attrs['class'][0] == 'not-detected--substances':
                    par = row.find('td').text.replace('Not detected:', '')
                    block = [strip_whitespace(x) for x in par.split(',')]
                    for analyte in block:
                        results.append({
                            'analysis': analysis,
                            'key': snake_case(analyte),
                            'name': analyte,
                            'value': 'ND',
                            'units': current_units,
                        })
    
    # Turn dates to ISO format.
    date_columns = [x for x in obs.keys() if x.startswith('date')]
    for date_column in date_columns:
        try:
            obs[date_column] = pd.to_datetime(obs[date_column]).isoformat()
        except:
            pass

    # Return the sample details with a new or re-minted `sample_id`.
    obs['analyses'] = list(set(analyses))
    obs['lab_results_url'] = url
    obs['results'] = results
    obs['sample_id'] = create_sample_id(
        private_key=json.dumps(obs['results']),
        public_key=obs['product_name'],
        salt=obs['producer'],
    )
    obs['coa_parsed_at'] = datetime.now().isoformat()
    return {**MCR_LABS, **obs}


def get_mcr_labs_samples(
        page_id: Any,
        cat: Optional[str] = 'all',
        order: Optional[str] = 'date-desc',
        search: Optional[str] = '',
        headers: Optional[Any] = None,
    ) -> list:
    """Get all test results from MCR Labs on a specific page.
    Args:
        page_id (str|int): The page number to get samples.
        cat (str): The category, `all` by default (optional). Options:
            `flower`, `concentrate`, `extract`, `mip`.
        order (str): The order to list results, `date-desc` by default
            (optional). Options: `date-desc`, `samplename`, `client`,
            `totalcann-desc`, `totalterp-desc`, `maxthc-desc`, `maxcbd-desc`.
        search (str): A particular search query.
        headers (dict): Headers for the HTTP request (optional).
    Returns:
        (list): A list of dictionaries of sample data.
    """
    # Get a page.
    base = MCR_LABS['url']
    url = f'{base}/ProductWeVeTested/AjaxSearch'
    params = {
        'category': cat,
        'order': order,
        'page': str(page_id),
        'searchString': search,
    }
    if headers is None:
        headers = DEFAULT_HEADERS
    response = requests.get(url, headers=headers, params=params)
    soup = BeautifulSoup(response.content, 'html.parser')

    # Get all of the products on the page.
    samples = []
    products = soup.find_all('li', attrs={'class': 'grid-item'})
    for product in products:

        # Get the sample details.
        obs = {}
        details = product.find('div', attrs={'class': 'reportTable'})

        # Get the product name.
        attrs = {'class': 'fth_name'}
        obs['product_name'] = details.find('div', attrs=attrs).text

        # Get the producer.
        attrs = {'class': 'fth_client'}
        obs['producer'] = details.find('div', attrs=attrs).text

        # Get the product type.
        # Optional: Standardize product types.
        attrs = {'class': 'fth_category'}
        obs['product_type'] = details.find('div', attrs=attrs).text

        # Get the total cannabinoids.
        attrs = {'class': 'fth_cannabinoids'}
        value = details.find('div', attrs=attrs).text
        obs['total_cannabinoids'] = strip_whitespace(value)

        # Get the total terpenes.
        attrs = {'class': 'fth_terpenes'}
        value = details.find('div', attrs=attrs).text
        obs['total_terpenes'] = strip_whitespace(value)

        # Get the date tested.
        try:
            obs['date_tested'] = format_iso_date(details.find('div', \
                attrs={'class': 'fth_date'}).text)
        except ValueError:
            print('Error parsing date:', obs)
            obs['date_tested'] = ''

        # Try to get the producer's URL.
        try:
            element = product.find('span', attrs={'class': 'url-linked'})
            href = element.attrs['data-url']
            obs['producer_url'] = '/'.join([base, href])
        except AttributeError:
            obs['producer_url'] = ''

        # Get the lab results URL.
        href = product.find('a')['href']
        obs['lab_results_url'] = '/'.join([base, href])

        # Get the image.
        image_url = product.find('img')['src']
        filename = image_url.split('/')[-1]
        obs['images'] = [{'url': image_url, 'filename': filename}]

        # Turn dates to ISO format.
        date_columns = [x for x in obs.keys() if x.startswith('date')]
        for date_column in date_columns:
            try:
                obs[date_column] = pd.to_datetime(obs[date_column]).isoformat()
            except:
                pass

        # Aggregate sample data.
        samples.append({**MCR_LABS, **obs})

    # Return the samples.
    return samples


def get_mcr_labs_test_results(
        starting_page: Optional[int] = 1,
        ending_page: Optional[int] = None,
        pause: Optional[float] = 3,
        verbose: Optional[bool] = True
    ) -> list:
    """Get all MCR Labs test results.
    Args:
        starting_page (int): The page to start collecting results,
            `1` by default (optional).
        ending_page (int): The page to end collecting results,
            `None` by default, which will collect to the end (optional).
        pause (float): The amount of time to wait between requests.
        verbose (bool): Whether to print out status, `True` by default
            (optional).
    Returns:
        (list): The complete sample data.
    """
    # Determine the pages to collect.
    page_count = get_mcr_labs_sample_count()
    total_pages = page_count['pages']
    if not ending_page:
        ending_page = total_pages
    if verbose:
        print('Getting samples for pages %i to %i.' % (starting_page, ending_page))
    samples = []

    # Iterate over all of the pages, index starting at 1.
    for page_id in range(starting_page, ending_page + 1):
        sample_data = get_mcr_labs_samples(page_id)
        samples += sample_data
        if page_id > 1 and page_id <= ending_page:
            sleep(pause)
    if verbose:
        print('Found %i samples.' % len(samples))

    # Get all of the sample details.
    # Optional: Log errors?
    rows = []
    for i, sample in enumerate(samples):
        try:
            lab_id = sample['lab_results_url'].split('/')[-1]
            details = get_mcr_labs_sample_details(None, lab_id)
            rows.append({**sample, **details})
            if i > 1:
                sleep(pause)
            if verbose:
                print('Collected sample:', lab_id)
        except:
            print('Failed to collect sample:', lab_id)
            continue

    # Return all of the sample data.
    return rows


def get_mcr_labs_producer_test_results():
    """Get all test results from MCR Labs for a specific producer.
    Returns:
        (list): A list of dictionaries of sample data.
    """
    # TODO: Implement.
    raise NotImplementedError


def get_mcr_labs_client_details():
    """Find details for any known producer of a given lab sample..
    Returns:
        (dict): The client's details data.
    """
    # TODO: Implement.
    raise NotImplementedError


def parse_mcrlabs_pdf(
        parser,
        doc: Any,
        coa_pdf: Optional[str] = '',
        **kwargs,
    ) -> dict:
    """Parse a MCR Labs COA PDF.
    Args:
        parser (CoADoc): A CoADoc parsing client.
        doc (str or PDF): A PDF file path or pdfplumber PDF.
        coa_pdf (str): A filename to use for the `coa_pdf` field (optional).
    Returns:
        (dict): The sample data, including:
            ✓ analyses
            ✓ batch_number
            ✓ methods
            - {analysis}_status TODO: Record the pass / fail status for each analysis.
            ✓ date_received
            ✓ date_tested
            ✓ lab_id
            ✓ metrc_lab_id
            ✓ metrc_source_id
            ✓ producer
            ✓ producer_street
            ✓ producer_city
            ✓ producer_state
            ✓ producer_zipcode
            ✓ producer_license_number
            ✓ product_name
            ✓ product_type
            ✓ results
            ✓ results_hash
            ✓ sample_hash
            ✓ serving_size
            ✓ sum_of_cannabinoids (Calculate simple sum of all cannabinoids)
            ✓ total_cannabinoids (Applying decarb rate to all acidic cannabinoids)
            ✓ total_cbd
            ✓ total_thc
            - total_terpenes
    """
    # Read the PDF.
    obs = {}
    if isinstance(doc, str):
        report = pdfplumber.open(doc)
        obs['coa_pdf'] = coa_pdf or doc.replace('\\', '/').split('/')[-1]
    else:
        report = doc
        obs['coa_pdf'] = coa_pdf or report.stream.name.replace('\\', '/').split('/')[-1]

    # Get the sample details.
    front_page = report.pages[0]
    tables = front_page.extract_tables()

    # Get the producer data.
    # FIXME: There is probably a more elegant way to parse this block.
    producer_details = tables[0][0][0].split('\n')
    if len(producer_details) == 5:
        producer = producer_details[1] + producer_details[2]
        address = ', '.join([producer_details[-2], producer_details[-1]])
    elif len(producer_details) == 6:
        producer = producer_details[1] + producer_details[2]
        address = ', '.join([
            producer_details[-3] + producer_details[-2],
            producer_details[-1]
        ])
    else:
        producer = producer_details[1]
        address = ', '.join([producer_details[-2], producer_details[-1]])

    # Optional: Get the producer address query for Google Places search.
    # query = tables[0][0][0].replace('\n', ' ').split('Client:')[-1] \
    #     .replace('  ', ' ').strip()

    # Get producer and producer's address.
    # FIXME: We may be able to look up the license with the MA Open Data API.
    address_parts = address.split(', ')
    state_parts = address_parts[-1].split(' ')
    obs['producer'] = producer
    obs['producer_street'] = address_parts[0]
    obs['producer_city'] = address_parts[1]
    obs['producer_zipcode'] = state_parts[-1]
    obs['producer_zipcode'] = state_parts[0].strip()

    # Lookup retailer for `cannabis_licenses` dataset.
    # FIXME: This may return multiple matches.
    # TODO: It may be better to lookup the license data with the Cannlytics API.
    ma_licenses = load_dataset('cannlytics/cannabis_licenses', 'ma')
    licenses = ma_licenses['data'].to_pandas()
    criterion = licenses['business_legal_name'].str.contains(obs['producer'])
    match = licenses.loc[criterion]
    if len(match) == 0:
        criterion = licenses['premise_street_address'].str.contains(obs['producer_street'])
        match = licenses.loc[criterion]
    if len(match) != 0:
        licensee = match.iloc[0]
        obs['producer_county'] = licensee['premise_county']
        obs['producer_latitude'] = licensee['premise_latitude']
        obs['producer_longitude'] = licensee['premise_longitude']
        obs['producer_license_number'] = licensee['license_number']
    
    # Note: Sample data differs on page 4:
    # "Serving size weight" 5,6,8 pages is "Sample Weight"
    while len(report.pages) < 5:
        try: 
          MCR_LABS_COA['fields']['Serving size weight'] = MCR_LABS_COA['fields'].pop('Sample Weight')
        except:
          break

    # Get the sample data.
    standard_fields = MCR_LABS_COA['fields']
    fields = tables[1][0]
    values = tables[1][-1]
    for i, field in enumerate(fields):
        key = field.replace('\n', '').strip()
        key = standard_fields.get(key, snake_case(key))
        value = values[i].replace('\n', '').strip()
        obs[key] = value

    # Get the Metrc source ID.
    # FIXME: Get both Metrc IDs if present.
    try:
        prefix = '1A40A01'
        front_page_text = front_page.extract_text()
        metrc_id = prefix + front_page_text.split(prefix)[-1].split('\n')[0].split('Date')[0]
        obs['metrc_source_id'] = metrc_id
        obs['metrc_lab_id'] = metrc_id
    except:
        obs['metrc_lab_id'] = None

    # Get the date reported to override `date_tested`.
    try:
        date = front_page_text.split('Report Date:')[-1].strip().split(' ')[:3]
        obs['date_tested'] = pd.to_datetime(', '.join(date))
    except:
        pass

    # Get and standardize the analyses.
    analyses_table = tables[-1][1:]
    analyses = [x[0] for x in analyses_table]
    analyses = [parser.analyses.get(x, x) for x in analyses]

    # Get the methods and results.
    # - analysis, units, key, name, value, mg_g, limit, status
    methods, results = [], []
    for page in report.pages[1:]:

        # Stop reading pages at the first quality control (QC) page.
        if (len(report.pages)) > 6 and (page.page_number > 5):
            break

        # Get the text and the tables.
        page_text = page.extract_text()
        tables = page.extract_tables()

        # Skip quality control (QC) pages.
        if 'Quality control checks' in page_text:
            continue

        # Get the analysis.
        analysis = page_text.split('Analyst:')[0].split('\n')[-1]
        analysis = analysis.split(' [')[0]
        analysis = parser.analyses.get(analysis, analysis)

        # Get the method.
        method = 'The sample' + page_text.split('Table ')[0] \
            .split('The sample')[-1].split('.')[0] + '.'
        methods.append(method)

        # Get the columns and units (for pesticide screening).
        if page.page_number == 5:
            analysis = 'pesticides'
            columns = [
                x.replace('Test Analysis', 'name') \
                 .replace('Result, ppb', 'value') \
                 .replace(' ppb', '') \
                 .replace('Limits', 'Limit') \
                 .replace('\n', '') for x in tables[0][0]
            ]
        
        # Get the columns and units (for terpene analysis).
        elif page.page_number == 6:
            analysis = "terpenes"
            columns = [
                x.replace('Terpene', 'name') \
                 .replace('Conc. (weight %)*', 'value') \
                 .replace('\n', '') for x in tables[0][0]
            ] 
        
        # Get the columns and units (for all other analyses).
        else:
            columns = [x.replace('\n', '') for x in tables[0][0]]
        column_text = ','.join(columns)
        columns = [re.sub('\(.*?\)|\[.*?\]', '', x).strip() for x in columns]
        columns = [standard_fields.get(x, x) for x in columns]
        if 'ppb' in column_text:
            units = 'ppb'
        else:
            units = STANDARD_UNITS.get(analysis)

        # Get the values from the tables.
        for index, table in enumerate(tables):

            # Skip the first table.
            table = table[1:]

            # Change analysis on heavy metal tables.
            if page.page_number == 4 and index == 1:
                tmp_analysis = page_text.split('Analyst:')[1].split('\n')[-1]
                tmp_analysis = tmp_analysis.split(' [')[0]
                tmp_analysis = parser.analyses.get(tmp_analysis, tmp_analysis)
                if tmp_analysis == 'heavy_metals':
                    analysis = tmp_analysis

            # Iterate over each row.
            for row in table:

                # Get the result data.
                result = {'analysis': analysis, 'units': units}
                total = False
                for i, cell in enumerate(row):

                    # Skip extra rows.
                    if i >= 7:
                        continue

                    # Get the result name and key.
                    column = columns[i]
                    if column == 'name':
                        name = cell.replace('\n', '').replace('  ', ' ')
                        key = snake_case(name)
                        result['key'] = parser.analytes.get(key, key)

                        # Record total THC and total CBD.
                        if cell.startswith('Total THC'):
                            obs['total_thc'] = convert_to_numeric(row[2])
                            total = True
                            continue
                        elif cell.startswith('Total CBD'):
                            obs['total_cbd'] = convert_to_numeric(row[2])
                            total = True
                            continue
                        elif key == 'test_analysis':
                            total = True
                            continue

                    # Handle multiple occurrences of value.
                    elif column == 'value':
                        serving_size = result.get('value', None)
                        if serving_size:
                            result['serving_size'] = serving_size
                    
                    # Skip nuisance values.
                    elif column == 'skip':
                        continue
                    
                    # Clean and record the value.
                    cell = cell.strip(' CFU/g')
                    cell = cell.strip(' in 1')
                    value = convert_to_numeric(cell)
                    if isinstance(value, str):
                        value = value.replace('\n', '') \
                            .replace('*', '') \
                            .replace('=', '') \
                            .replace('<100', '99') \
                            .replace('%', '') \
                            .replace('Not Detected', 'ND') \
                            .replace('  ', ' ')
                    result[column] = value
                
                # Record the result, skipping total rows.
                if not total:
                    results.append(result)

    # Close the report.
    report.close()

    # Calculate total (active) cannabinoids and sum of cannabinoids.
    sum_of_cannabinoids, total_cannabinoids = 0, 0
    for result in results:
        if result['analysis'] == 'cannabinoids':
            value = result['value']
            if isinstance(value, float):
                if result['key'].endswith('a'):
                    total_cannabinoids += DECARB * value
                else:
                    total_cannabinoids += value
                sum_of_cannabinoids += value
    obs['total_cannabinoids'] = total_cannabinoids
    obs['sum_of_cannabinoids'] = sum_of_cannabinoids

    # Hot-fix: MCR Labs reports total mycotoxins.
    # for i, result in enumerate(results):
    #     if result['analysis'] == 'mycotoxins':
    #         result['key'] = 'total_aflatoxins'
    #         results.append(result)
    #         del results[i]
    #         break

    # Turn dates to ISO format.
    date_columns = [x for x in obs.keys() if x.startswith('date')]
    for date_column in date_columns:
        try:
            obs[date_column] = pd.to_datetime(obs[date_column]).isoformat()
        except:
            pass

    # Finish data collection with a freshly minted sample ID.
    obs = {**MCR_LABS, **obs}
    obs['analyses'] = json.dumps(analyses)
    obs['coa_algorithm_version'] = __version__
    obs['coa_parsed_at'] = datetime.now().isoformat()
    obs['methods'] = json.dumps(methods)
    obs['results'] = results
    obs['results_hash'] = create_hash(results)
    obs['sample_id'] = create_sample_id(
        private_key=json.dumps(results),
        public_key=obs['product_name'],
        salt=obs.get('producer', obs.get('date_tested', 'cannlytics.eth')),
    )
    # Note: Fix for `TypeError: Object of type float32 is not JSON serializable`.
    obs['sample_hash'] = create_hash(str(obs))
    return obs


def parse_mcrlabs_coa(
        parser,
        doc: Any,
        **kwargs,
    ) -> dict:
    """Parse a MCR Labs COA PDF or URL.
    Args:
        doc (str or PDF): A PDF file path or pdfplumber PDF.
        kwargs (arguments): Arguments to pass to the parsing algorithms.
    Returns:
        (dict): The sample data.
    """
    if isinstance(doc, str):
        if doc.endswith('.pdf'):
            data = parse_mcrlabs_pdf(parser, doc, **kwargs)
        else:
            data = get_mcr_labs_sample_details(parser, doc, **kwargs)
    else:
        data = parse_mcrlabs_pdf(parser, doc, **kwargs)
    if isinstance(doc, str):
        data['coa_pdf'] = doc.replace('\\', '/').split('/')[-1]
    elif isinstance(doc, pdfplumber.pdf.PDF):
        data['coa_pdf'] = doc.stream.name.replace('\\', '/').split('/')[-1]
    return data


# === Tests ===
# Notes: Uncomment tests to perform them.
# Checked tests have been successfully performed by Cannlytics.
# Contact: <admin@cannlytics.com>.
if __name__ == '__main__':

    from cannlytics.data.coas import CoADoc
    from cannlytics.utils.utils import to_excel_with_style
    from datetime import datetime
    import pandas as pd

    # Specify where your test data lives.
    DATA_DIR = '../../../.datasets/lab_results'

    # [✓] TEST: Get the total number of samples.
    # page_count = get_mcr_labs_sample_count(per_page=30)
    # assert page_count is not None

    # [✓] TEST: Get the samples on a given page.
    # samples = get_mcr_labs_samples('6')
    # assert samples is not None

    # [✓] TEST: Get a sample's details.
    # sample = get_mcr_labs_sample_details(None, 'rooted-labs-distillate_2')
    # assert sample is not None

    # [✓] TEST: Get an infused products's results (`product_type == 'mip'`).
    # details = get_mcr_labs_sample_details(None, '67545')
    # assert details is not None

    # [✓] TEST: Get a range of the samples.
    # print('Getting range of the samples.')
    # all_results = get_mcr_labs_test_results(
    #     starting_page=150,
    #     ending_page=250,
    #     pause=0.2,
    # )
    # timestamp = datetime.now().isoformat()[:19].replace(':', '-')
    # datafile = f'{DATA_DIR}/mcr-lab-results-{timestamp}.xlsx'
    # to_excel_with_style(pd.DataFrame(all_results), datafile)
    # print('Finished getting range of samples.')

    # [✓] TEST: Get only the most recent samples.
    # print('Getting the most recent samples.')
    # recent_results = get_mcr_labs_test_results(ending_page=2)
    # timestamp = datetime.now().isoformat()[:19].replace(':', '-')
    # datafile = f'{DATA_DIR}/mcr-lab-results-{timestamp}.xlsx'
    # to_excel_with_style(pd.DataFrame(recent_results), datafile)
    # print('Tested getting the most recent samples.')

    # [✓] TEST: Parse a MCR Labs COA PDF from 2022.
    # parser = CoADoc()
    # doc = '../../../tests/assets/coas/mcr-labs/Bulk 100mg Milk Chocolate Drops.pdf'
    # lab = parser.identify_lims(doc, lims={'MCR Labs': MCR_LABS})
    # assert lab == 'MCR Labs'
    # data = parse_mcrlabs_pdf(parser, doc)
    # data = parse_mcrlabs_coa(parser, doc)
    # assert data is not None
    # print('Parsed:', doc)

    # [✓] TEST: Parse a MCR Labs COA PDF from 2021.
    # parser = CoADoc()
    # doc = '../../../tests/assets/coas/mcr-labs/Watermelon 30mg Gummy.pdf'
    # data = parse_mcrlabs_pdf(parser, doc)
    # assert data is not None
    # print('Parsed:', doc)

    # [✓] Test: Parse and save MCR Labs COAs observed in the wild.
    # parser = CoADoc()
    # docs = [
    #     '../../../tests/assets/coas/mcr-labs/mcr-labs-ma-4pgs.pdf',
    #     '../../../tests/assets/coas/mcr-labs/mcr-labs-ma-5pgs.pdf',
    #     '../../../tests/assets/coas/mcr-labs/mcr-labs-ma-6pgs.pdf',
    #     '../../../tests/assets/coas/mcr-labs/mcr-labs-ma-8pgs.pdf',
    # ]
    # for doc in docs:
    #     data = parse_mcrlabs_coa(parser, doc)
    #     assert data is not None
    #     print('Parsed:', doc)
    #     outfile = doc.replace('.pdf', '.xlsx')
    #     parser.save(data, outfile)



"""
Calculation Model | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Calculation data model.
"""
# Standard imports.
from dataclasses import dataclass

# Internal imports.
from .base import Model


@dataclass
class Calculation(Model):
    """A calculation is applied to measurements to determine final
    results, such as applying dilution factor, etc."""
    _collection = 'organizations/%s/calculations'
    formula: str = ''
